#!/usr/bin/env python3
# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import sys
import shutil
from pathlib import Path
import subprocess

import venv

out = Path(__file__).parent / "for-buildozer"
out.mkdir(exist_ok=True)

buildozer_venv = Path(__file__).parent / "venv-buildozer"

contrib = Path(__file__).parent.parent

with (out / "buildozer.spec").open("w") as buildozerspec:
    buildozerspec.write("""\
[app]
title = aiocoap widget demo
package.name = aiocoap_widget_demo
package.domain = org.example
source.dir = .
source.include_exts = py
version = 0.1

# docutils for kivy rst widget
#
# everything after that is aiocoap[prettyprint,oscore,ws] spelled out (without
# aiocoap as that's copied in below). for some reason, ge25519's dependency on
# fe25519 needs to be spelled out.
requirements = python3,kivy,docutils,termcolor,cbor2,pygments,cbor-diag,cryptography,filelock,ge25519,fe25519,websockets
orientation = portrait

fullscreen = 0
android.permissions = INTERNET

# You may need to adjust this to your device
android.arch = armeabi-v7a
""")

shutil.copy(contrib / "aiocoap-kivy-widget", out / "main.py")
# We could use aiocoap from PyPI, but so far things only work with this
# branch's version.
shutil.copytree(contrib / "widgets_common", out / "widgets_common", dirs_exist_ok=True)
shutil.copytree(contrib.parent / "aiocoap", out / "aiocoap", dirs_exist_ok=True)

venv.create(buildozer_venv, with_pip=True)
subprocess.check_call(
    [
        "bash",
        "-e",
        "-c",
        """\
        source ../venv-buildozer/bin/activate
        pip install cython buildozer
        buildozer $@
""",
        "buildozer",
    ]
    + sys.argv[1:],
    cwd=out,
)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Integration tests for EDHOC"""

import unittest
import tempfile
from pathlib import Path

import aiocoap

from .test_server import TestServerBase, WithClient, WithTestServer
from .fixtures import no_warnings


class WithEdhocPair(WithTestServer, WithClient):
    server_style: str
    client_style: str

    client_knows_server = True
    server_knows_client = True

    use_combined_edhoc = True

    def setUp(self):
        super().setUp()

        # Unlike in DTLS, we do these as a pair because we generate both key
        # pairs and distribute them to both sides

        from aiocoap import edhoc
        import aiocoap.oscore_sitewrapper

        self.tmpdir = tempfile.TemporaryDirectory()
        tmpbase = Path(self.tmpdir.name)

        serverkey_file = tmpbase / "server.cosekey"
        serverkey = edhoc.CoseKeyForEdhoc.generate(serverkey_file)
        serverccs = serverkey.as_ccs(kid=bytes.fromhex("00"), subject="s")
        if self.client_knows_server:
            serverccs_for_client = serverccs
        else:
            serverccs_for_client = {"unauthenticated": True}

        clientkey_file = tmpbase / "client.cosekey"
        clientkey = edhoc.CoseKeyForEdhoc.generate(clientkey_file)
        clientccs = clientkey.as_ccs(kid=bytes.fromhex("01"), subject="c")
        if self.server_knows_client:
            clientccs_for_server = clientccs
        else:
            clientccs_for_server = {"unauthenticated": True}

        # FIXME shouldn't that only need the servernamealias?
        client_entry = {
            "edhoc-oscore": {
                "suite": 2,
                "method": 3,
                "use_combined_edhoc": self.use_combined_edhoc,
                "own_cred_style": self.client_style,
                "own_cred": clientccs,
                "private_key_file": str(clientkey_file),
                "peer_cred": serverccs_for_client,
            }
        }
        self.client.client_credentials.load_from_dict(
            {"coap://%s/*" % self.servernamealias: client_entry}
        )
        self.client.client_credentials.load_from_dict(
            {"coap://%s/*" % self.servernetloc: client_entry}
        )

        self.server.server_credentials.load_from_dict(
            {
                "coap://%s/*" % self.servernetloc: {
                    "edhoc-oscore": {
                        "suite": 2,
                        "method": 3,
                        "own_cred_style": self.server_style,
                        "own_cred": serverccs,
                        "private_key_file": str(serverkey_file),
                    }
                },
                ":edhocclient": {
                    "edhoc-oscore": {
                        "suite": 2,
                        "method": 3,
                        "peer_cred": clientccs_for_server,
                    }
                },
            }
        )

        self.server.serversite = aiocoap.oscore_sitewrapper.OscoreSiteWrapper(
            self.server.serversite, self.server.server_credentials
        )

    def tearDown(self):
        super().tearDown()
        self.tmpdir.cleanup()


edhoc_modules = aiocoap.defaults.oscore_missing_modules()


@unittest.skipIf(edhoc_modules, "EDHOC/OSCORE missing modules (%s)" % (edhoc_modules,))
class BaseServerEdhoc(TestServerBase, WithEdhocPair):
    @no_warnings
    def test_whoami_is_client(self):
        request = self.build_request()
        request.opt.uri_path = ["whoami"]
        response = self.fetch_response(request)
        self.assertTrue(
            b":edhocclient" in response.payload,
            f"Expected to see own role in response, got {response.payload=}",
        )


class TestServerEdhocKidKid(BaseServerEdhoc):
    server_style = "by-key-id"
    client_style = "by-key-id"


class TestServerEdhocValueValue(BaseServerEdhoc):
    server_style = "by-value"
    client_style = "by-value"


class TestServerEdhocAnonKid(BaseServerEdhoc):
    server_style = "by-value"
    client_style = "by-key-id"
    client_knows_server = False


class TestServerEdhocKidAnon(BaseServerEdhoc):
    server_style = "by-key-id"
    client_style = "by-value"
    server_knows_client = False


class TestServerEdhocVeryVerbose(TestServerEdhocValueValue):
    use_combined_edhoc = False


# that's not supposed to be tested, its child classes are
del BaseServerEdhoc



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tests for the aiocoap.oscore module based on the test vectors."""

import unittest

import aiocoap

oscore_modules = aiocoap.defaults.oscore_missing_modules()

if not oscore_modules:
    import aiocoap.oscore

    # shortcut definition, as this will be used all over the place with copy-pasted
    # values from the specification
    h = bytes.fromhex

    C1_KEY = h("0102030405060708090a0b0c0d0e0f10")
    C1_SALT = h("9e7ca92223786340")

    C2_KEY = h("0102030405060708090a0b0c0d0e0f10")
    C2_SALT = None

    C3_KEY = h("0102030405060708090a0b0c0d0e0f10")
    C3_SALT = h("9e7ca92223786340")
    C3_ID_CTX = h("37cbf3210017a2d3")

    default_algorithm = aiocoap.oscore.AES_CCM_16_64_128()
    default_hashfun = aiocoap.oscore.hashfunctions["sha256"]

    import aiocoap.oscore

    class NonsavingSecurityContext(
        aiocoap.oscore.CanProtect,
        aiocoap.oscore.CanUnprotect,
        aiocoap.oscore.SecurityContextUtils,
    ):
        def post_seqnoincrease(self):
            # obviously, don't use this anywhere else, especially not with secret
            # keys
            pass


_skip_unless_oscore = unittest.skipIf(
    oscore_modules, "Modules missing for running OSCORE tests: %s" % (oscore_modules,)
)


@_skip_unless_oscore
class TestOSCOAPStatic(unittest.TestCase):
    def test_c1_1(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b""
        secctx.recipient_id = b"\x01"
        secctx.id_context = None
        secctx.derive_keys(C1_SALT, C1_KEY)

        # info not compared; that would be tricky to extract and adds no value

        self.assertEqual(
            secctx.sender_key,
            h("f0910ed7295e6ad4b54fc793154302ff"),
            "Sender key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.recipient_key,
            h("ffb14e093c94c9cac9471648b4f98710"),
            "Recipient key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.common_iv,
            h("4622d4dd6d944168eefb54987c"),
            "Common IV key derivation disagrees with test vector",
        )

        sender_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.sender_id, default_algorithm
        )
        self.assertEqual(
            sender_nonce_0,
            h("4622d4dd6d944168eefb54987c"),
            "Sender nonce disagrees with test vector",
        )
        recipient_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.recipient_id, default_algorithm
        )
        self.assertEqual(
            recipient_nonce_0,
            h("4722d4dd6d944169eefb54987c"),
            "Recipient nonce disagrees with test vector",
        )

    def test_c1_2(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b"\x01"
        secctx.recipient_id = b""
        secctx.id_context = None
        secctx.derive_keys(C1_SALT, C1_KEY)

        # info not compared; that would be tricky to extract and adds no value

        self.assertEqual(
            secctx.sender_key,
            h("ffb14e093c94c9cac9471648b4f98710"),
            "Sender key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.recipient_key,
            h("f0910ed7295e6ad4b54fc793154302ff"),
            "Recipient key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.common_iv,
            h("4622d4dd6d944168eefb54987c"),
            "Common IV key derivation disagrees with test vector",
        )

        sender_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.sender_id, default_algorithm
        )
        self.assertEqual(
            sender_nonce_0,
            h("4722d4dd6d944169eefb54987c"),
            "Sender nonce disagrees with test vector",
        )
        recipient_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.recipient_id, default_algorithm
        )
        self.assertEqual(
            recipient_nonce_0,
            h("4622d4dd6d944168eefb54987c"),
            "Recipient nonce disagrees with test vector",
        )

    def test_c2_1(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b"\x00"
        secctx.recipient_id = b"\x01"
        secctx.id_context = None
        secctx.derive_keys(C2_SALT, C2_KEY)

        # info not compared; that would be tricky to extract and adds no value

        self.assertEqual(
            secctx.sender_key,
            h("321b26943253c7ffb6003b0b64d74041"),
            "Sender key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.recipient_key,
            h("e57b5635815177cd679ab4bcec9d7dda"),
            "Recipient key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.common_iv,
            h("be35ae297d2dace910c52e99f9"),
            "Common IV key derivation disagrees with test vector",
        )

        sender_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.sender_id, default_algorithm
        )
        self.assertEqual(
            sender_nonce_0,
            h("bf35ae297d2dace910c52e99f9"),
            "Sender nonce disagrees with test vector",
        )
        recipient_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.recipient_id, default_algorithm
        )
        self.assertEqual(
            recipient_nonce_0,
            h("bf35ae297d2dace810c52e99f9"),
            "Recipient nonce disagrees with test vector",
        )

    # skipping the server side for c.2.2 as it is very redundant

    def test_c3_1(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b""
        secctx.recipient_id = b"\x01"
        secctx.id_context = C3_ID_CTX
        secctx.derive_keys(C3_SALT, C3_KEY)

        # info not compared; that would be tricky to extract and adds no value

        self.assertEqual(
            secctx.sender_key,
            h("af2a1300a5e95788b356336eeecd2b92"),
            "Sender key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.recipient_key,
            h("e39a0c7c77b43f03b4b39ab9a268699f"),
            "Recipient key derivation disagrees with test vector",
        )
        self.assertEqual(
            secctx.common_iv,
            h("2ca58fb85ff1b81c0b7181b85e"),
            "Common IV key derivation disagrees with test vector",
        )

        sender_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.sender_id, default_algorithm
        )
        self.assertEqual(
            sender_nonce_0,
            h("2ca58fb85ff1b81c0b7181b85e"),
            "Sender nonce disagrees with test vector",
        )
        recipient_nonce_0 = secctx._construct_nonce(
            b"\0", secctx.recipient_id, default_algorithm
        )
        self.assertEqual(
            recipient_nonce_0,
            h("2da58fb85ff1b81d0b7181b85e"),
            "Recipient nonce disagrees with test vector",
        )

    def test_c4(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b""
        secctx.recipient_id = b"\x01"
        secctx.id_context = None
        secctx.derive_keys(C1_SALT, C1_KEY)
        secctx.sender_sequence_number = 20

        unprotected = aiocoap.Message.decode(
            h("44015d1f00003974396c6f63616c686f737483747631")
        )
        outer_message, _ = secctx.protect(unprotected)
        outer_message.mid = unprotected.mid
        outer_message.token = unprotected.token
        outer_message.mtype = unprotected.mtype

        # again skipping some intermediary values that are all checked in the final result as well

        encoded = outer_message.encode()
        self.assertEqual(
            encoded,
            h("44025d1f00003974396c6f63616c686f7374620914ff612f1092f1776f1c1668b3825e"),
            "Encoded message differs",
        )

    def test_c5(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b"\x00"
        secctx.recipient_id = b"\x01"
        secctx.id_context = None
        secctx.derive_keys(C2_SALT, C2_KEY)
        secctx.sender_sequence_number = 20

        unprotected = aiocoap.Message.decode(
            h("440171c30000b932396c6f63616c686f737483747631")
        )
        outer_message, _ = secctx.protect(unprotected)
        outer_message.mid = unprotected.mid
        outer_message.token = unprotected.token
        outer_message.mtype = unprotected.mtype

        # again skipping some intermediary values that are all checked in the final result as well

        encoded = outer_message.encode()
        self.assertEqual(
            encoded,
            h(
                "440271c30000b932396c6f63616c686f737463091400ff4ed339a5a379b0b8bc731fffb0"
            ),
            "Encoded message differs",
        )

    def test_c6(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b""
        secctx.recipient_id = b"\x01"
        secctx.id_context = C3_ID_CTX
        secctx.derive_keys(C3_SALT, C3_KEY)
        secctx.sender_sequence_number = 20

        unprotected = aiocoap.Message.decode(
            h("44012f8eef9bbf7a396c6f63616c686f737483747631")
        )
        outer_message, _ = secctx.protect(unprotected, kid_context=True)
        outer_message.mid = unprotected.mid
        outer_message.token = unprotected.token
        outer_message.mtype = unprotected.mtype

        # again skipping some intermediary values that are all checked in the final result as well

        encoded = outer_message.encode()
        # FIXME: This is composed from the expected ciphertext, not the protected coap request, see https://github.com/core-wg/oscoap/issues/241
        self.assertEqual(
            encoded,
            h(
                "44022f8eef9bbf7a396c6f63616c686f73746b19140837cbf3210017a2d3ff72cd7273fd331ac45cffbe55c3"
            ),
            "Encoded message differs",
        )

    def test_c7(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b"\x01"
        secctx.recipient_id = b""
        secctx.id_context = None
        secctx.derive_keys(C1_SALT, C1_KEY)

        unprotected = aiocoap.Message.decode(
            h("64455d1f00003974ff48656c6c6f20576f726c6421")
        )
        request_sender_id = secctx.recipient_id
        request_piv_short = b"\x14"
        request_nonce = secctx._construct_nonce(
            request_piv_short, request_sender_id, default_algorithm
        )
        outer_message, _ = secctx.protect(
            unprotected,
            aiocoap.oscore.RequestIdentifiers(
                request_sender_id, request_piv_short, True, aiocoap.POST
            ),
        )
        outer_message.mid = unprotected.mid
        outer_message.token = unprotected.token
        outer_message.mtype = unprotected.mtype

        # again skipping some intermediary values that are all checked in the final result as well

        encoded = outer_message.encode()
        self.assertEqual(
            encoded,
            h("64445d1f0000397490ffdbaad1e9a7e7b2a813d3c31524378303cdafae119106"),
            "Encoded message differs",
        )

    def test_c8(self):
        secctx = NonsavingSecurityContext()
        secctx.alg_aead = default_algorithm
        secctx.hashfun = default_hashfun
        secctx.sender_id = b"\x01"
        secctx.recipient_id = b""
        secctx.id_context = None
        secctx.derive_keys(C1_SALT, C1_KEY)
        secctx.sender_sequence_number = 0

        unprotected = aiocoap.Message.decode(
            h("64455d1f00003974ff48656c6c6f20576f726c6421")
        )
        request_sender_id = secctx.recipient_id
        request_piv_short = b"\x14"
        request_nonce = secctx._construct_nonce(
            request_piv_short, request_sender_id, default_algorithm
        )
        outer_message, _ = secctx.protect(
            unprotected,
            aiocoap.oscore.RequestIdentifiers(
                request_sender_id, request_piv_short, False, aiocoap.POST
            ),
        )
        outer_message.mid = unprotected.mid
        outer_message.token = unprotected.token
        outer_message.mtype = unprotected.mtype

        # again skipping some intermediary values that are all checked in the final result as well

        encoded = outer_message.encode()
        self.assertEqual(
            encoded,
            h("64445d1f00003974920100ff4d4c13669384b67354b2b6175ff4b8658c666a6cf88e"),
            "Encoded message differs",
        )


@_skip_unless_oscore
class TestOSCOAAsymmetric(unittest.TestCase):
    """Test asymmetric algorithms

    As the Group OSCORE document currently has no test vectors, this is using
    values from the IETF109 hackathon.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # From https://github.com/ace-wg/Hackathon-109/blob/master/GroupKeys.md,
        # "Rikard Test 1"

        self.r1_1_d = bytes.fromhex(
            "FEA2190084748436543C5EC8E329D2AFBD7068054F595CA1F987B9E43E2205E6"
        )
        self.r1_1_y = bytes.fromhex(
            "64CE3DD128CC4EFA6DE209BE8ABD111C7272F612C2DB654057B6EC00FBFB0684"
        )
        self.r1_1_x = bytes.fromhex(
            "1ADB2AB6AF48F17C9877CF77DB4FA39DC0923FBE215E576FE6F790B1FF2CBC96"
        )

        self.r1_2_d = bytes.fromhex(
            "DA2593A6E0BCC81A5941069CB76303487816A2F4E6C0F21737B56A7C90381597"
        )
        self.r1_2_y = bytes.fromhex(
            "1897A28666FE1CC4FACEF79CC7BDECDC271F2A619A00844FCD553A12DD679A4F"
        )
        self.r1_2_x = bytes.fromhex(
            "0EB313B4D314A1001244776D321F2DD88A5A31DF06A6EEAE0A79832D39408BC1"
        )

        self.r1_3_d = bytes.fromhex(
            "BF31D3F9670A7D1342259E700F48DD9983A5F9DF80D58994C667B6EBFD23270E"
        )
        self.r1_3_y = bytes.fromhex(
            "5694315AD17A4DA5E3F69CA02F83E9C3D594712137ED8AFB748A70491598F9CD"
        )
        self.r1_3_x = bytes.fromhex(
            "FAD4312A45F45A3212810905B223800F6CED4BC8D5BACBC8D33BB60C45FC98DD"
        )

        # "Rikard Test 2"
        self.r2_csalg = -8
        self.r2_csalg_params = [[1], [1, 6]]

        self.r2_1_private = bytes.fromhex(
            "397CEB5A8D21D74A9258C20C33FC45AB152B02CF479B2E3081285F77454CF347"
        )
        self.r2_1_public = bytes.fromhex(
            "CE616F28426EF24EDB51DBCEF7A23305F886F657959D4DF889DDFC0255042159"
        )

        self.r2_2_private = bytes.fromhex(
            "70559B9EECDC578D5FC2CA37F9969630029F1592AFF3306392AB15546C6A184A"
        )
        self.r2_2_public = bytes.fromhex(
            "2668BA6CA302F14E952228DA1250A890C143FDBA4DAED27246188B9E42C94B6D"
        )

        self.r2_3_private = bytes.fromhex(
            "E550CD532B881D52AD75CE7B91171063E568F2531FBDFB32EE01D1910BCF810F"
        )
        self.r2_3_public = bytes.fromhex(
            "5394E43633CDAC96F05120EA9F21307C9355A1B66B60A834B53E9BF60B1FB7DF"
        )

        # from https://github.com/ace-wg/Hackathon-109/blob/master/GroupDerivation.md

        self.r1_shared_12 = bytes.fromhex(
            "56ede6c59e919031cfc8afa3e74a7b7615c2e7a08494cf3638c78757293adc80"
        )
        self.r1_shared_13 = bytes.fromhex(
            "f568ec5f7df45db137fc79a27595eba737b62e8ee385c7309e316dd409de6953"
        )

        # Not actually used; these are already verified in tests_util_cryptography's vectors
        self.r2_shared_12 = bytes.fromhex(
            "4546babdb9482396c167af11d21953bfa49eb9f630c45de93ee4d3b9ef059576"
        )
        self.r2_shared_13 = bytes.fromhex(
            "bb11648af3dfebb35e612914a7a21fc751b001aceb0267c5536528e2b9261450"
        )

    def alg(self):
        alg_sign = aiocoap.oscore.Ed25519()
        alg_pairwise = aiocoap.oscore.Ed25519()
        self.assertEqual(alg_sign.value, self.r2_csalg)
        return (alg_sign, alg_pairwise)

    def test_publickey_derivation(self):
        alg, _ = self.alg()
        self.assertEqual(self.r2_1_public, alg.public_from_private(self.r2_1_private))
        self.assertEqual(self.r2_2_public, alg.public_from_private(self.r2_2_private))
        self.assertEqual(self.r2_3_public, alg.public_from_private(self.r2_3_private))

    def _test_keypair(self, alg_sign, alg_pairwise, private, public):
        body = b""
        aad = b""
        signature = alg_sign.sign(body, aad, private)
        alg_sign.verify(signature, body, aad, public)

        self.assertRaises(
            aiocoap.oscore.ProtectionInvalid,
            lambda: alg_sign.verify(signature, body + b"x", aad, public),
        )

    @unittest.skip
    def test_publickey_signatures(self):
        alg = self.alg()

        self._test_keypair(*alg, self.r2_1_private, self.r2_1_public)

        self._test_keypair(*alg, self.r2_2_private, self.r2_2_public)
        self._test_keypair(*alg, self.r2_3_private, self.r2_3_public)

    @unittest.skip
    def test_generation(self):
        alg = self.alg()

        for alg in aiocoap.oscore.algorithms_staticstatic.values():
            random_key = alg.generate()
            public_key = alg.public_from_private(random_key)

            self._test_keypair(*alg, random_key, public_key)

            second_random = alg.generate()
            second_public = alg.public_from_private(second_random)
            self.assertEqual(
                alg.staticstatic(random_key, second_public),
                alg.staticstatic(second_random, public_key),
            )

    def test_ecdsa_vectors(self):
        alg = aiocoap.oscore.ECDSA_SHA256_P256()

        r1_1 = alg.from_private_parts(self.r1_1_x, self.r1_1_y, self.r1_1_d)
        r1_2 = alg.from_private_parts(self.r1_2_x, self.r1_2_y, self.r1_2_d)
        r1_3 = alg.from_private_parts(self.r1_3_x, self.r1_3_y, self.r1_3_d)

        self.assertEqual(
            alg.staticstatic(r1_1, alg.public_from_private(r1_2)), self.r1_shared_12
        )
        self.assertEqual(
            alg.staticstatic(r1_3, alg.public_from_private(r1_1)), self.r1_shared_13
        )


@_skip_unless_oscore
class TestOSCORECompression(unittest.TestCase):
    def compare_uncompress(
        self, ref_option, ref_payload, ref_protected, ref_unprotected, ref_ciphertext
    ):
        message = aiocoap.Message(payload=ref_payload, oscore=ref_option)
        protected_serialized, protected, unprotected, ciphertext = (
            aiocoap.oscore.CanUnprotect._extract_encrypted0(message)
        )

        self.assertEqual(protected, ref_protected, "Protected dictionary mismatch")
        self.assertEqual(
            unprotected, ref_unprotected, "Unprotected dictionary mismatch"
        )
        self.assertEqual(ciphertext, ref_ciphertext, "Ciphertext mismatch")

    def compare_compress(
        self, ref_option, ref_payload, ref_protected, ref_unprotected, ref_ciphertext
    ):
        option, payload = aiocoap.oscore.CanProtect._compress(
            ref_protected, ref_unprotected, ref_ciphertext
        )

        self.assertEqual(option, ref_option, "Compressed option mismatch")
        self.assertEqual(payload, ref_payload, "Compressed payload mismatch")

    def compare_all(
        self, ref_option, ref_payload, ref_protected, ref_unprotected, ref_ciphertext
    ):
        self.compare_uncompress(
            ref_option, ref_payload, ref_protected, ref_unprotected, ref_ciphertext
        )
        self.compare_compress(
            ref_option, ref_payload, ref_protected, ref_unprotected, ref_ciphertext
        )

    def test_empty(self):
        self.compare_all(b"", b"1234", {}, {}, b"1234")

    def test_short(self):
        self.compare_all(
            b"\x01\x00", b"1234", {}, {aiocoap.oscore.COSE_PIV: b"\x00"}, b"1234"
        )

    def test_long(self):
        self.compare_all(
            b"\x05ABCDE", b"1234", {}, {aiocoap.oscore.COSE_PIV: b"ABCDE"}, b"1234"
        )

    def test_kid(self):
        self.compare_all(
            b"\x0bABC--------",
            b"1234",
            {},
            {aiocoap.oscore.COSE_PIV: b"ABC", aiocoap.oscore.COSE_KID: b"--------"},
            b"1234",
        )

    def test_idcontext(self):
        self.compare_all(
            b"\x1bABC\x03abc--------",
            b"1234",
            {},
            {
                aiocoap.oscore.COSE_PIV: b"ABC",
                aiocoap.oscore.COSE_KID: b"--------",
                aiocoap.oscore.COSE_KID_CONTEXT: b"abc",
            },
            b"1234",
        )

    def test_idcontext_nopiv(self):
        self.compare_all(
            b"\x18\x03abc--------",
            b"1234",
            {},
            {
                aiocoap.oscore.COSE_KID: b"--------",
                aiocoap.oscore.COSE_KID_CONTEXT: b"abc",
            },
            b"1234",
        )

    def test_counterisgnature(self):
        # This is only as correct as it gets with the interactions between
        # determining the countersignature (or its length) and uncompression:
        # The flag is registered, but its value is deferred to a later step
        # (that'd actually know the algorithm) and relies on the later process
        # to move data over from the plaintext
        self.compare_all(
            b"\x2bABC--",
            b"1234sigsigsig",
            {},
            {
                aiocoap.oscore.COSE_PIV: b"ABC",
                aiocoap.oscore.COSE_KID: b"--",
                aiocoap.oscore.COSE_COUNTERSIGNATURE0: aiocoap.oscore.PRESENT_BUT_NO_VALUE_YET,
            },
            b"1234sigsigsig",
        )

    def test_reserved(self):
        self.assertRaises(
            aiocoap.oscore.DecodeError,
            lambda: self.compare_uncompress(b"\x4bABC--", b"1234", None, None, None),
        )
        self.assertRaises(
            aiocoap.oscore.DecodeError,
            lambda: self.compare_uncompress(b"\x8bABC--", b"1234", None, None, None),
        )

    def test_unknown(self):
        self.assertRaises(
            RuntimeError, lambda: self.compare_compress(None, None, {"x": "y"}, {}, b"")
        )
        self.assertRaises(
            RuntimeError, lambda: self.compare_compress(None, None, {}, {"x": "y"}, b"")
        )


@_skip_unless_oscore
class TestAlgorithms(unittest.TestCase):
    def test_aescbc(self):
        # Test vectors generated using PyCryptodome as described at
        # https://codeigo.com/python/aes-encryption-and-decryption-in-python-64-128-256/
        # ("Implementation of AES-CBC Encryption and Decryption in Python" with
        # they key length set to 16)

        from aiocoap.oscore import A128CBC

        key = b"3\x8b\xd0\xf2\x18&\x15\nS\x14%\xec!\x16\n$"
        iv = b"7\x07\xe3TA65~\x8d%'&\x8e\x1anE"
        plain = b"This is the message to be encrypted"
        expected_ciphertext = b"\x05\xbb\xd0\xea\xc2\x97pvF-P\x9a\x9e\xa4=\x94\xac\xad\x00\xc0/\x00&\x7f\x83\xca@c\xeew\xc3\x16\x14h\xca\x023\xb6Vq\xf4\x8d\xb7\xc2S3\xfa\xae"

        aad = b""

        encrypted = A128CBC.encrypt(plain, aad, key, iv)
        self.assertEqual(encrypted, expected_ciphertext)
        decrypted = A128CBC.decrypt(encrypted, aad, key, iv)
        self.assertEqual(decrypted, plain)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Module in which all meaningful numbers are collected. Most of the submodules
correspond to IANA registries.

The contents of the :mod:`.constants`, :mod:`.types` and :mod:`.codes` modules
are accessible through this module directly; :mod:`.contentformat`'s and
:mod:`.optionnumbers`' sole :class:`.contentformat.ContentFormat` and
:class:`.optionnumbers.OptionNumber` classes are accessible in the same way.
"""

import warnings
import string

from . import constants, types, codes
from .contentformat import ContentFormat, _MediaTypes, _MediaTypesRev
from .optionnumbers import OptionNumber

# These lists are hard-coded in because static checkers need them explicit. The code below helps keep it up to date.
# fmt: off
from .constants import COAPS_PORT, COAP_PORT, MAX_REGULAR_BLOCK_SIZE_EXP, MCAST_ALL, MCAST_IPV4_ALLCOAPNODES, MCAST_IPV6_LINKLOCAL_ALLCOAPNODES, MCAST_IPV6_LINKLOCAL_ALLNODES, MCAST_IPV6_SITELOCAL_ALLCOAPNODES, MCAST_IPV6_SITELOCAL_ALLNODES, SHUTDOWN_TIMEOUT, TransportTuning
from .types import Type, CON, NON, ACK, RST
from .codes import Code, EMPTY, GET, POST, PUT, DELETE, FETCH, PATCH, iPATCH, CREATED, DELETED, VALID, CHANGED, CONTENT, CONTINUE, BAD_REQUEST, UNAUTHORIZED, BAD_OPTION, FORBIDDEN, NOT_FOUND, METHOD_NOT_ALLOWED, NOT_ACCEPTABLE, REQUEST_ENTITY_INCOMPLETE, CONFLICT, PRECONDITION_FAILED, REQUEST_ENTITY_TOO_LARGE, UNSUPPORTED_CONTENT_FORMAT, UNPROCESSABLE_ENTITY, TOO_MANY_REQUESTS, INTERNAL_SERVER_ERROR, NOT_IMPLEMENTED, BAD_GATEWAY, SERVICE_UNAVAILABLE, GATEWAY_TIMEOUT, PROXYING_NOT_SUPPORTED, HOP_LIMIT_REACHED, CSM, PING, PONG, RELEASE, ABORT
__all__ = ['COAPS_PORT', 'COAP_PORT', 'MAX_REGULAR_BLOCK_SIZE_EXP', 'MCAST_ALL', 'MCAST_IPV4_ALLCOAPNODES', 'MCAST_IPV6_LINKLOCAL_ALLCOAPNODES', 'MCAST_IPV6_LINKLOCAL_ALLNODES', 'MCAST_IPV6_SITELOCAL_ALLCOAPNODES', 'MCAST_IPV6_SITELOCAL_ALLNODES', 'SHUTDOWN_TIMEOUT', 'TransportTuning', 'Type', 'CON', 'NON', 'ACK', 'RST', 'Code', 'EMPTY', 'GET', 'POST', 'PUT', 'DELETE', 'FETCH', 'PATCH', 'iPATCH', 'CREATED', 'DELETED', 'VALID', 'CHANGED', 'CONTENT', 'CONTINUE', 'BAD_REQUEST', 'UNAUTHORIZED', 'BAD_OPTION', 'FORBIDDEN', 'NOT_FOUND', 'METHOD_NOT_ALLOWED', 'NOT_ACCEPTABLE', 'REQUEST_ENTITY_INCOMPLETE', 'CONFLICT', 'PRECONDITION_FAILED', 'REQUEST_ENTITY_TOO_LARGE', 'UNSUPPORTED_CONTENT_FORMAT', 'UNPROCESSABLE_ENTITY', 'TOO_MANY_REQUESTS', 'INTERNAL_SERVER_ERROR', 'NOT_IMPLEMENTED', 'BAD_GATEWAY', 'SERVICE_UNAVAILABLE', 'GATEWAY_TIMEOUT', 'PROXYING_NOT_SUPPORTED', 'HOP_LIMIT_REACHED', 'CSM', 'PING', 'PONG', 'RELEASE', 'ABORT', 'OptionNumber', 'ContentFormat']
# fmt: on

if __debug__:
    _generated_all = (
        constants.__all__
        + types.__all__
        + codes.__all__
        + ["OptionNumber", "ContentFormat"]
    )
    if _generated_all != __all__:
        warnings.warn(f"""Hardcoded __all__ is out of sync (as are imports, probably), please updated to

        from .constants import {", ".join(constants.__all__)}
        from .types import {", ".join(types.__all__)}
        from .codes import {", ".join(codes.__all__)}
        __all__ = {_generated_all}""")

media_types = _MediaTypes()
media_types_rev = _MediaTypesRev()


def __getattr__(name):
    if name[0] in string.ascii_uppercase and hasattr(
        constants._default_transport_tuning, name
    ):
        warnings.warn(
            f"{name} is deprecated, use through the message's transport_tuning instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return getattr(constants._default_transport_tuning, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This tests launch the command line utility aiocoap-client in a sub-process.

The aiocoap-proxy utility is tested in test_proxy inside this process as
orchestration of success reporting is not that easy with a daemon process;
aiocoap-rd might need to get tested in a similar way to -proxy."""

import subprocess
import unittest
import os

import aiocoap.defaults

from .test_server import WithTestServer, no_warnings, asynctest
from .common import PYTHON_PREFIX

linkheader_modules = aiocoap.defaults.linkheader_missing_modules()

AIOCOAP_CLIENT = PYTHON_PREFIX + ["./aiocoap-client"]
AIOCOAP_RD = PYTHON_PREFIX + ["./aiocoap-rd"]


class TestCommandlineClient(WithTestServer):
    @no_warnings
    def test_help(self):
        # CI environments often don't have any locale set. That's not
        # representative of the interactive environments that run --help
        # (though it may be for others). Setting C.UTF-8 because at least pypy3
        # 7.3 doesn't default to a UTF-8 enabled mode, and the help output is
        # not just ASCII.
        helptext = subprocess.check_output(
            AIOCOAP_CLIENT + ["--help"], env={"LANG": "C.UTF-8"}
        )
        self.assertTrue(helptext.startswith(b"usage: aiocoap-client "))

    @no_warnings
    @asynctest
    async def test_get(self):
        await self.loop.run_in_executor(None, self._test_get)

    def _test_get(self):
        # FIXME style: subprocesses could be orchestrated using asyncio as well
        empty_default = subprocess.check_output(
            AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty"]
        )
        self.assertEqual(empty_default, b"")

        empty_json = subprocess.check_output(
            AIOCOAP_CLIENT
            + [
                "coap://" + self.servernetloc + "/empty",
                "--accept",
                "application/json",
                "--quiet",
            ]
        )
        self.assertEqual(empty_json, b"{}")

        verbose = subprocess.check_output(
            AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty", "-v"],
            stderr=subprocess.STDOUT,
        )
        # It'd not be actually wrong to have info level messages in here, but
        # they should at least not start appearing unnoticed.
        self.assertEqual(
            verbose, b"", "Unexpected info-level messages in simple request"
        )

        debug = subprocess.check_output(
            AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty", "-v", "-v"],
            stderr=subprocess.STDOUT,
        )
        self.assertTrue(
            b"DEBUG:coap:Incoming message" in debug,
            "Not even some (or unexpected) output in aiocoap-client -vv",
        )

        quiet = subprocess.check_output(
            AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty", "--quiet"],
            stderr=subprocess.STDOUT,
        )
        self.assertEqual(quiet, b"")

        explicit_code = subprocess.check_output(
            AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty", "-m1"]
        )
        self.assertEqual(explicit_code, b"")

        if not aiocoap.defaults.prettyprint_missing_modules():
            json_formatted = subprocess.check_output(
                AIOCOAP_CLIENT
                + [
                    "coap://" + self.servernetloc + "/answer",
                    "--accept",
                    "application/json",
                    "--pretty-print",
                ]
            )
            # Concrete formatting may vary, but it should be indented
            self.assertEqual(
                json_formatted.replace(b"\r", b""), b'{\n    "answer": 42\n}'
            )

            json_colorformatted = subprocess.check_output(
                AIOCOAP_CLIENT
                + [
                    "coap://" + self.servernetloc + "/answer",
                    "--accept",
                    "application/json",
                    "--pretty-print",
                    "--color",
                ]
            )
            self.assertTrue(
                b"\x1b[" in json_colorformatted,
                "No color indication in pretty-printed JSON",
            )
            self.assertTrue(
                b"    " in json_colorformatted,
                "No indentation in color-pretty-printed JSON",
            )

            json_coloronly = subprocess.check_output(
                AIOCOAP_CLIENT
                + [
                    "coap://" + self.servernetloc + "/answer",
                    "--accept",
                    "application/json",
                    "--color",
                ]
            )
            self.assertTrue(
                b"\x1b[" in json_coloronly, "No color indication in color-printed JSON"
            )
            self.assertTrue(
                b"    " not in json_coloronly, "Indentation in color-printed JSON"
            )

            cbor_formatted = subprocess.check_output(
                AIOCOAP_CLIENT
                + [
                    "coap://" + self.servernetloc + "/answer",
                    "--accept",
                    "application/cbor",
                    "--pretty-print",
                ]
            )
            # Concrete formatting depends on cbor-diag package
            self.assertEqual(cbor_formatted, b'{"answer": 42}')

    @no_warnings
    @asynctest
    async def test_post(self):
        await self.loop.run_in_executor(None, self._test_post)

    def _test_post(self):
        replace_foo = subprocess.check_output(
            AIOCOAP_CLIENT
            + [
                "coap://" + self.servernetloc + "/replacing/one",
                "-m",
                "post",
                "--payload",
                "f00",
            ]
        )
        self.assertEqual(replace_foo, b"fOO")

        replace_file = subprocess.check_output(
            AIOCOAP_CLIENT
            + [
                "coap://" + self.servernetloc + "/replacing/one",
                "-m",
                "post",
                "--payload",
                "@" + os.devnull,
            ]
        )
        self.assertEqual(replace_file, b"")

    @no_warnings
    @asynctest
    async def test_erroneous(self):
        await self.loop.run_in_executor(None, self._test_erroneous)

    def _test_erroneous(self):
        with self.assertRaises(subprocess.CalledProcessError):
            # non-existant method
            subprocess.check_output(
                AIOCOAP_CLIENT + ["coap://" + self.servernetloc + "/empty", "-mSPAM"],
                stderr=subprocess.STDOUT,
            )

        with self.assertRaises(subprocess.CalledProcessError):
            # not a URI
            subprocess.check_output(
                AIOCOAP_CLIENT + ["coap::://" + self.servernetloc + "/empty"],
                stderr=subprocess.STDOUT,
            )

        with self.assertRaises(subprocess.CalledProcessError):
            # relative URI
            subprocess.check_output(
                AIOCOAP_CLIENT + ["/empty"], stderr=subprocess.STDOUT
            )

        with self.assertRaises(subprocess.CalledProcessError):
            # non-existant mime type
            subprocess.check_output(
                AIOCOAP_CLIENT
                + ["coap://" + self.servernetloc + "/empty", "--accept", "spam/eggs"],
                stderr=subprocess.STDOUT,
            )

        try:
            # No full URI given
            subprocess.check_output(
                AIOCOAP_CLIENT + [self.servernetloc + "/empty"],
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            self.assertTrue(
                "URL incomplete: Must start with a scheme." in e.output.decode("utf8")
            )
            # It must also show the extra_help
            self.assertTrue(
                "Most URLs in aiocoap need to be given with a scheme"
                in e.output.decode("utf8")
            )
        else:
            raise AssertionError(
                "Calling aiocoap-client without a full URI should fail."
            )

        try:
            subprocess.check_output(
                AIOCOAP_CLIENT + ["http://" + self.servernetloc + "/empty"],
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            self.assertTrue(
                "No remote endpoint set for request" in e.output.decode("utf8")
            )
            # Extra help even gives concrete output
            self.assertTrue(
                f"The message is set up for use with a proxy (because the scheme of 'http://{self.servernetloc}/empty' is not supported)"
                in e.output.decode("utf8")
            )
        else:
            raise AssertionError(
                "Calling aiocoap-client without a HTTP URI should fail."
            )

    @no_warnings
    @asynctest
    async def test_noproxy(self):
        await self.loop.run_in_executor(None, self._test_noproxy)

    def _test_noproxy(self):
        # Having this successful and just return text is a bespoke weirdness of
        # the /empty resource (and MultiRepresentationResource in general).
        # Once https://github.com/chrysn/aiocoap/issues/268 is resolved, their
        # workarounds that make this not just ignore the critical proxy options
        # in the first place will go away, and this will need to process
        # aiocoap-client failing regularly.
        stdout = subprocess.check_output(
            AIOCOAP_CLIENT
            + [
                "coap://0.0.0.0/empty",
                "--proxy",
                "coap://" + self.servernetloc,
                "--verbose",
            ],
            stderr=subprocess.STDOUT,
        )

        self.assertEqual(stdout, b"This is no proxy")


class TestCommandlineRD(unittest.TestCase):
    @unittest.skipIf(
        linkheader_modules,
        "Modules missing for running RD tests: %s" % (linkheader_modules,),
    )
    def test_help(self):
        helptext = subprocess.check_output(AIOCOAP_RD + ["--help"])
        self.assertTrue(helptext.startswith(b"usage: aiocoap-rd "))



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import unittest

import aiocoap

from . import common
from .test_server import TestServer, WithClient, WithTestServer

PSK = "the PSK key"
identity = "test-client"


class WithDTLSClient(WithClient):
    def setUp(self):
        super().setUp()
        # FIXME shouldn't that only need the servernamealias?
        self.client.client_credentials.load_from_dict(
            {
                "coaps://%s/*" % self.servernamealias: {
                    "dtls": {
                        "psk": {"ascii": PSK},
                        "client-identity": {"ascii": identity},
                    }
                }
            }
        )
        self.client.client_credentials.load_from_dict(
            {
                "coaps://%s/*" % self.servernetloc: {
                    "dtls": {
                        "psk": {"ascii": PSK},
                        "client-identity": {"ascii": identity},
                    }
                }
            }
        )


class WithDTLSServer(WithTestServer):
    def setUp(self):
        super().setUp()
        self.server.server_credentials.load_from_dict(
            {
                ":client": {
                    "dtls": {
                        "psk": {"ascii": PSK},
                        "client-identity": {"ascii": identity},
                    }
                }
            }
        )


dtls_modules = aiocoap.defaults.dtls_missing_modules()


@unittest.skipIf(
    dtls_modules or common.dtls_disabled,
    "DTLS missing modules (%s) or disabled in this environment" % (dtls_modules,),
)
class TestServerDTLS(TestServer, WithDTLSClient, WithDTLSServer):
    # as with TestServerTCP

    def build_request(self):
        request = super().build_request()
        # odd default port
        request.requested_scheme = "coaps"
        return request


del TestServer

if __name__ == "__main__":
    # due to the imports, you'll need to run this as `python3 -m tests.test_server`
    common.run_fixture_as_standalone_server(WithDTLSServer)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Tests for the file server.

This expects a file server at a CoAP server's root directory to be empty; files
are created, and different ways of triggering conflict responses are triggered
(eg. by overwriting a file, and then attempting to overwrite it with a
different one expecting the old state).
"""

import argparse
import asyncio
import os
import sys
import tempfile
import unittest

import aiocoap
from aiocoap import Message
from aiocoap.numbers.codes import *
from aiocoap.util import hostportjoin

from .common import PYTHON_PREFIX, CapturingSubprocess
from .test_server import WithAsyncLoop, WithClient, asynctest

SERVER_NETLOC = hostportjoin("::1", None)
AIOCOAP_FILESERVER = PYTHON_PREFIX + [
    "./aiocoap-fileserver",
    "--write",
    "--bind",
    SERVER_NETLOC,
]


@unittest.skipIf(
    aiocoap.defaults.linkheader_missing_modules(),
    "Module missing for running fileserver tests: %s"
    % (aiocoap.defaults.linkheader_missing_modules(),),
)
class WithFileServer(WithAsyncLoop):
    def setUp(self):
        super().setUp()
        ready = self.loop.create_future()
        self.__done = self.loop.create_future()

        self.filedir = tempfile.mkdtemp(suffix="-fileserver")

        self.__task = self.loop.create_task(self.run_server(ready, self.__done))
        self.loop.run_until_complete(ready)

    # This might be overly complex; it was stripped down from the more intricate OSCORE plug tests
    async def run_server(self, readiness, done):
        self.process, process_outputs = await self.loop.subprocess_exec(
            CapturingSubprocess,
            *(AIOCOAP_FILESERVER + ["-vvvvvvvv"]),
            self.filedir,
            stdin=None,
        )

        while True:
            if b"Server ready to receive requests" in process_outputs.stderr:
                break
            if self.process.get_returncode() is not None:
                readiness.set_exception(
                    RuntimeError(
                        "File server process terminated during startup:\n%s\n%s"
                        % (
                            process_outputs.stdout.decode("utf8"),
                            process_outputs.stderr.decode("utf8"),
                        )
                    )
                )
                return
            await process_outputs.read_more
        readiness.set_result(True)

        while True:
            if self.process.get_returncode() is not None:
                break
            await process_outputs.read_more

        done.set_result(
            (
                self.process.get_returncode(),
                process_outputs.stdout,
                process_outputs.stderr,
            )
        )

        self.process.close()

    def tearDown(self):
        # Don't leave this over, even if anything is raised during teardown
        self.process.terminate()

        # Checking whether it's empty
        os.rmdir(self.filedir)

        super().tearDown()

        code, out, err = self.loop.run_until_complete(self.__done)


class TestFileServer(WithFileServer, WithClient):
    @asynctest
    async def test_fullcycle(self):
        await work_fileserver(
            self.client, "coap://%s/" % SERVER_NETLOC, self.assertTrue
        )


# Implemented as a single function to ease use from the command line
async def work_fileserver(ctx, base_uri, assert_):
    assert_(
        base_uri.endswith("/") and base_uri.count("/") == 3,
        "Base URI needs to be of shape coapsomething://hostname/ "
        "(with precisely these slashes)",
    )

    req = Message(code=GET, uri=base_uri)
    res = await ctx.request(req).response_raising
    assert_(res.code == CONTENT)
    assert_(res.opt.content_format == 40, "Directory listing is not in link-format")
    assert_(res.payload == b"", "Directory is initially not empty")

    file1_body = b"Hello World\n" * 200
    req = Message(code=PUT, uri=base_uri + "file", payload=file1_body)
    res = await ctx.request(req).response_raising
    assert_(res.code == CHANGED)
    early_etag = res.opt.etag
    assert_(early_etag is not None, "PUT did not already send an ETag")

    # TBD: Could persist a content format

    req = Message(code=GET, uri=base_uri + "file")
    res = await ctx.request(req).response_raising
    assert_(res.code == CONTENT)
    assert_(res.payload == file1_body)
    etag1 = res.opt.etag
    assert_(etag1 == early_etag, "PUT response ETag is not GET response ETag")
    assert_(etag1 is not None, "No ETag returned")

    # Revalidate
    req = Message(code=GET, uri=base_uri + "file", etags=[b"synthetic", etag1])
    res = await ctx.request(req).response_raising
    assert_(res.code == VALID)
    assert_(not res.payload)
    assert_(res.opt.etag == etag1)

    file2_body = b"It is diffrent now."
    req = Message(
        code=PUT, uri=base_uri + "file", payload=file2_body, if_none_match=True
    )
    res = await ctx.request(req).response
    assert_(
        res.code == PRECONDITION_FAILED,
        "Overwrite succeeded even though expected empty",
    )

    req = Message(
        code=PUT, uri=base_uri + "file", payload=file2_body, if_match=[b"synthetic"]
    )
    res = await ctx.request(req).response
    assert_(
        res.code == PRECONDITION_FAILED,
        "Overwrite succeeded even though expected to conflict",
    )

    req = Message(code=PUT, uri=base_uri + "file", payload=file2_body, if_match=[etag1])
    res = await ctx.request(req).response_raising
    assert_(res.code == CHANGED)

    # Empty ETag is actually illegal, but the workaround to force the server to
    # send one is even cruder, see
    # https://github.com/core-wg/corrclar/issues/46
    #
    # This is not needed with the original body because that is blockwise'd.
    req = Message(code=GET, uri=base_uri + "file", etag=b"")
    res = await ctx.request(req).response_raising
    assert_(res.code == CONTENT)
    assert_(res.payload == file2_body)
    etag2 = res.opt.etag
    assert_(etag2 is not None, "No ETag returned")

    req = Message(code=DELETE, uri=base_uri + "file", if_match=[b"not the one"])
    res = await ctx.request(req).response
    assert_(
        res.code == PRECONDITION_FAILED,
        "Deletion should not pass with a non-matching ETag",
    )

    # Actual cleanup
    req = Message(code=DELETE, uri=base_uri + "file", if_match=[etag2])
    res = await ctx.request(req).response_raising
    assert_(res.code == DELETED)

    # Just to be sure
    req = Message(code=DELETE, uri=base_uri + "file")
    res = await ctx.request(req).response
    assert_(res.code == NOT_FOUND)  # DELETED might be OK too under idempotency rules?


async def run_standalone():
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument("base_uri")
    args = p.parse_args()

    ctx = await aiocoap.Context.create_client_context()

    def assert_(truth, error=""):
        if not truth:
            print("Assertion failed:", error)
            sys.exit(1)

    await work_fileserver(ctx, args.base_uri, assert_)


if __name__ == "__main__":
    # due to the imports, you'll need to run this as `python3 -m tests.test_fileserver`
    asyncio.run(run_standalone())



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import logging
from aiocoap import oscore_sitewrapper
import aiocoap.resource as resource
import aiocoap
import asyncio

logging.basicConfig(level=logging.INFO)
logging.getLogger("coap-server").setLevel(logging.DEBUG)


async def main():
    # Resource tree creation
    root = resource.Site()

    root.add_resource(
        [".well-known", "core"], resource.WKCResource(root.get_resources_as_linkheader)
    )

    server_credentials = aiocoap.credentials.CredentialsMap()

    root = oscore_sitewrapper.OscoreSiteWrapper(root, server_credentials)

    protocol = await aiocoap.Context.create_server_context(root)

    # Keys from IETF109 plug test: Rikard Test 2 Entity 1
    server_credentials[":a"] = aiocoap.oscore.SimpleGroupContext(
        algorithm=aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
        hashfun=aiocoap.oscore.hashfunctions[aiocoap.oscore.DEFAULT_HASHFUNCTION],
        alg_signature=aiocoap.oscore.Ed25519(),
        alg_group_enc=aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
        alg_pairwise_key_agreement=aiocoap.oscore.EcdhSsHkdf256(),
        group_id=bytes.fromhex("DD11"),
        master_secret=bytes.fromhex("11223344556677889900AABBCCDDEEFF"),
        master_salt=bytes.fromhex("1F2E3D4C5B6A7081"),
        sender_id=bytes.fromhex("0A"),
        private_key=bytes.fromhex(
            "397CEB5A8D21D74A9258C20C33FC45AB152B02CF479B2E3081285F77454CF347"
        ),
        peers={
            bytes.fromhex("51"): bytes.fromhex(
                "2668BA6CA302F14E952228DA1250A890C143FDBA4DAED27246188B9E42C94B6D"
            ),
            bytes.fromhex("52"): bytes.fromhex(
                "5394E43633CDAC96F05120EA9F21307C9355A1B66B60A834B53E9BF60B1FB7DF"
            ),
            bytes.fromhex("dc"): aiocoap.oscore.DETERMINISTIC_KEY,
        },
    )

    await asyncio.get_running_loop().create_future()


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This tests advanced cases of blockwise transfer; simple sequential transfers
are covered in test_server.TestServer.test_replacing_resource."""

import asyncio

import unittest
import aiocoap
import aiocoap.defaults

from .test_server import (
    WithTestServer,
    WithClient,
    no_warnings,
    asynctest,
    BigResource,
    BasicTestingSite,
)


class BigChunkyResource(BigResource):
    async def render_get(self, request):
        request.remote.maximum_block_size_exp = 3
        return await super().render_get(request)


class ChunkyTestingSite(BasicTestingSite):
    def __init__(self):
        super().__init__()

        self.add_resource(["big", "chunky"], BigChunkyResource())


class WithChunkyTestServer(WithTestServer):
    TestingSite = ChunkyTestingSite


class TestBlockwise(WithChunkyTestServer, WithClient):
    # tracked as https://github.com/chrysn/aiocoap/issues/58; behavior can be successful more or less by chance
    @unittest.skip
    @no_warnings
    @asynctest
    async def test_sequential(self):
        """Test whether the client serializes simultaneous block requests"""

        pattern1 = b"01234 first pattern" + b"01" * 1024
        pattern2 = b"01234 second pattern" + b"02" * 1024

        request1 = aiocoap.Message(
            uri="coap://" + self.servernetloc + "/replacing/one",
            code=aiocoap.POST,
            payload=pattern1,
        )
        request2 = aiocoap.Message(
            uri="coap://" + self.servernetloc + "/replacing/one",
            code=aiocoap.POST,
            payload=pattern2,
        )

        responses = []
        for response in asyncio.as_completed(
            [self.client.request(r).response for r in [request1, request2]]
        ):
            response = await response
            self.assertTrue(
                response.code.is_successful(),
                "Simultaneous blockwise requests caused error.",
            )
            responses.append(response.payload)

        self.assertSetEqual(
            set(responses), set(x.replace(b"0", b"O") for x in (pattern1, pattern2))
        )

    @no_warnings
    @asynctest
    async def test_client_hints(self):
        """Test whether a handle_blockwise=True request takes a block2 option
        set in it as a hint to start requesting with a low size right away

        That way of hinting at size requests is not documented, but understood
        and occasionally used; the test primarily serves to identify problems
        the server has with initially low block sizes."""

        resp = await self.client.request(
            aiocoap.Message(
                uri="coap://" + self.servernetloc + "/big",
                block2=(0, 0, 3),
                code=aiocoap.GET,
            )
        ).response

        self.assertEqual(resp.code, aiocoap.CONTENT, "Request was unsuccessful")
        self.assertEqual(
            self._count_received_messages(),
            (len(resp.payload) + 127) // 128,
            "Response not chunked into 128 bytes",
        )

    @no_warnings
    @asynctest
    async def test_server_hints(self):
        """Test whether the needs_blockwise server mechanism considers size
        exponent limits of the remote.

        Mutating the remote as it is done in the BigChunkyResource is not
        documented, but helps verify that the information in
        maximum_block_size_exp is generally used."""

        resp = await self.client.request(
            aiocoap.Message(
                uri="coap://" + self.servernetloc + "/big/chunky",
                code=aiocoap.GET,
            )
        ).response

        self.assertEqual(resp.code, aiocoap.CONTENT, "Request was unsuccessful")
        self.assertEqual(
            self._count_received_messages(),
            (len(resp.payload) + 127) // 128,
            "Response not chunked into 128 bytes",
        )

    @no_warnings
    @asynctest
    async def test_client_hints_block1(self):
        """Test whether a client can successfully indicate per-remote sizes
        even for the block1 phase."""

        # same as BigResource's content
        payload = b"0123456789----------" * 512
        reqmsg = aiocoap.Message(
            uri="coap://" + self.servernetloc + "/replacing/one",
            code=aiocoap.PUT,
            payload=payload,
        )
        reqmsg.remote.maximum_block_size_exp = 3
        resp = await self.client.request(reqmsg).response

        self.assertEqual(resp.code, aiocoap.CHANGED, "Request was unsuccessful")
        self.assertEqual(
            self._count_received_messages(),
            (len(payload) + 127) // 128,
            "Response not chunked into 128 bytes",
        )

    _received_logmsg = "Incoming message <aiocoap.Message at"

    def _count_received_messages(self):
        # only client-side received empty-acks are counted
        return sum(
            self._received_logmsg in x.msg
            for x in self.handler
            if x.name != "coap-server"
        )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import asyncio
import aiocoap
import errno

from .test_server import WithTestServer, WithClient, no_warnings, asynctest


class TestClientWithSetHost(WithTestServer, WithClient):
    set_uri_host = True

    @no_warnings
    @asynctest
    async def test_uri_parser(self):
        request = aiocoap.Message(code=aiocoap.GET)
        request_uri = "coap://" + self.servernetloc + "/empty?query=a&query=b"
        request.set_request_uri(request_uri, set_uri_host=self.set_uri_host)
        self.assertEqual(
            request.get_request_uri(),
            request_uri,
            "Request URL does not round-trip in request",
        )
        response = await self.client.request(request).response
        self.assertEqual(
            response.get_request_uri(),
            request_uri,
            "Request URL does not round-trip in response",
        )
        self.assertEqual(response.code, aiocoap.CONTENT, "Request URL building failed")

        request = aiocoap.Message(code=aiocoap.GET)
        request.set_request_uri(
            "coap://" + self.servernamealias + "/empty", set_uri_host=self.set_uri_host
        )
        self.assertEqual(
            request.get_request_uri(), "coap://" + self.servernamealias + "/empty"
        )
        response = await self.client.request(request).response
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Resolving WithTestServer.servernamealias failed",
        )
        if self.set_uri_host:
            self.assertEqual(
                response.get_request_uri(),
                "coap://" + self.servernamealias + "/empty",
                "Host name did not get round-tripped",
            )
        else:
            # The simple6 transport misreports remotes to which a socket was
            # opened with a name.
            if "simple6" not in list(
                aiocoap.defaults.get_default_clienttransports(loop=self.loop)
            ):
                self.assertEqual(
                    response.get_request_uri(),
                    "coap://" + self.servernetloc + "/empty",
                    "Response's request URI is not numeric in hostname-less query",
                )

    @no_warnings
    @asynctest
    async def test_uri_parser2(self):
        """A difficult test because it is prone to keeping the transport
        around, bothering later tests"""

        request = aiocoap.Message(code=aiocoap.GET)
        request.set_request_uri(
            "coap://" + self.servernetloc + ":9999/empty",
            set_uri_host=self.set_uri_host,
        )
        resp = self.client.request(request).response
        try:
            # give the request some time to finish getaddrinfo
            result = await asyncio.as_completed([resp], timeout=0.1).__next__()
        except aiocoap.error.NetworkError as e:
            # This is a bit stricter than what the API indicates, but hey, we
            # can still relax the tests.
            self.assertTrue(isinstance(e.__cause__, OSError))
            # ECONNREFUSED: linux; ECONNRESET: win32
            self.assertTrue(e.__cause__.errno in (errno.ECONNREFUSED, errno.ECONNRESET))
        except asyncio.TimeoutError:
            self.fail(
                "Request to non-opened port did not come back with 'Connection Refused' immediately"
            )
        else:
            self.fail(
                "Request to non-opened port did not come back with 'Connection Refused', but another result: %s"
                % (result,)
            )
        self.assertTrue(
            request.remote.hostinfo.endswith(":9999"), "Remote port was not parsed"
        )
        resp.cancel()


class TestClientWithHostlessMessages(TestClientWithSetHost):
    set_uri_host = False


class TestClientOther(WithTestServer, WithClient):
    @no_warnings
    # can't do @asynctest because of assertRaises
    def test_raising(self):
        """This test obtains results via the response_raising property of a
        Request."""
        yieldfrom = self.loop.run_until_complete

        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://" + self.servernetloc + "/empty"
        )
        response = yieldfrom(self.client.request(request).response_raising)
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Response access via response_raising failed",
        )

        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://" + self.servernetloc + "/nonexistent"
        )
        ## @FIXME i'd like to assertRaises(NotFound), see docstring of
        # :class:`ResponseWrappingError`
        self.assertRaises(
            aiocoap.error.ResponseWrappingError,
            yieldfrom,
            self.client.request(request).response_raising,
        )

    @no_warnings
    @asynctest
    async def test_nonraising(self):
        """This test obtains results via the response_nonraising property of a
        Request."""
        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://" + self.servernetloc + "/empty"
        )
        response = await self.client.request(request).response_nonraising
        self.assertEqual(
            response.code,
            aiocoap.CONTENT,
            "Response access via response_nonraising failed",
        )

        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://cant.resolve.this.example./empty"
        )
        response = await self.client.request(request).response_nonraising
        self.assertEqual(response.code, aiocoap.INTERNAL_SERVER_ERROR)

    @no_warnings
    @asynctest
    async def test_freeoncancel(self):
        # As there's no programmatic feedback about what actually gets sent,
        # looking at the logs is the easiest option, even though it will
        # require occasional adjustment when logged messages change.
        #
        # FIXME Currently, this *only* checks for whether later responses are
        # rejected, it does *not* check for whether the response runner is
        # freed as well (primarily because that'd need _del_to_be_sure to be
        # useable in an async context).

        # With immediate cancellation, nothing is sent. Note that we don't
        # ensure this per documentation, but still it's good to see when this
        # changes.
        loglength = len(self.handler.list)
        request = aiocoap.Message(
            code=aiocoap.GET, uri="coap://" + self.servernetloc + "/empty"
        )
        self.resp = self.client.request(request).response
        self.resp.cancel()
        self.assertEqual(
            loglength,
            len(self.handler.list),
            "Something was logged during request creation and immediate cancellation: %r"
            % (self.handler.list[loglength:],),
        )

        # FIXME: What is slightly weird here is that on uvloop, something
        # *does* get sent -- but it appears that that happens even after the
        # cancellation. That behavior oly gets apparent when test_freeoncancel
        # and test_freeoncancel_non were run in a single test, because then the
        # ACK Content from the prior test also got counted.

    @no_warnings
    @asynctest
    async def test_freeoncancel_non(self):
        # With a NON, the response should take long. (Not trying to race the
        # "I'm taking too long"-ACK by making the sleep short enough).
        # Note that the resource implementation deliberately sends responses as CON,
        # as to allow us to peek into the internals of aiocoap by
        # looking at wehter it returns a RST or an ACK.
        request = aiocoap.Message(
            code=aiocoap.GET,
            uri="coap://" + self.servernetloc + "/slow",
            mtype=aiocoap.NON,
        )
        self.resp = self.client.request(request).response
        # Wait for the request to actually be sent
        while not any("Sending request" in l.getMessage() for l in self.handler.list):
            await asyncio.sleep(0.001)
        # Now the request was sent, let's look at what happens during and after the cancellation
        loglength = len(self.handler.list)
        self.resp.cancel()
        await asyncio.sleep(0.4)  # server takes 0.2 to respond
        logmsgs = self.handler.list[loglength:]
        unmatched_msgs = [
            l
            for l in logmsgs
            if "could not be matched to any request" in l.getMessage()
        ]
        self.assertEqual(
            len(unmatched_msgs), 1, "The incoming response was not treated as unmatched"
        )



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module assists in creating OSCORE servers by proving a wrapper around a
:class:aiocoap.resource.Site. It enforces no access control, but just indicates
to the resources whether a client is authenticated by setting the request's
remote property adaequately.
"""

import logging
from typing import Optional
import uuid
import io

import cbor2
import lakers

import aiocoap
from aiocoap import interfaces
from aiocoap import oscore, error
import aiocoap.pipe
from .numbers.codes import FETCH, POST
from .numbers.optionnumbers import OptionNumber
from . import edhoc

from aiocoap.transports.oscore import OSCOREAddress


class OscoreSiteWrapper(interfaces.Resource):
    def __init__(self, inner_site, server_credentials):
        self.log = logging.getLogger("coap-server.oscore-site")

        self._inner_site = inner_site
        self.server_credentials = server_credentials

    def get_resources_as_linkheader(self):
        # Not applying any limits while WKCResource does not either
        #
        # Not adding `;osc` everywhere as that is excessive (and not telling
        # much, as one won't know *how* to get those credentials)
        return self._inner_site.get_resources_as_linkheader()

    async def render(self, request):
        raise RuntimeError(
            "OscoreSiteWrapper can only be used through the render_to_pipe interface"
        )

    async def needs_blockwise_assembly(self, request):
        raise RuntimeError(
            "OscoreSiteWrapper can only be used through the render_to_pipe interface"
        )

    async def render_to_pipe(self, pipe):
        request = pipe.request

        if request.opt.uri_path == (".well-known", "edhoc"):
            # We'll have to take that explicitly, otherwise we'd need to rely
            # on a resource to be prepared by the user in the site with a
            # cyclical reference closed after site construction
            await self._render_edhoc_to_pipe(pipe)
            return

        try:
            unprotected = oscore.verify_start(request)
        except oscore.NotAProtectedMessage:
            # ie. if no object_seccurity present
            await self._inner_site.render_to_pipe(pipe)
            return

        if request.code not in (FETCH, POST):
            raise error.MethodNotAllowed

        try:
            sc = self.server_credentials.find_oscore(unprotected)
        except KeyError:
            if request.mtype == aiocoap.CON:
                raise error.Unauthorized("Security context not found")
            else:
                return

        try:
            unprotected, seqno = sc.unprotect(request)
        except error.RenderableError as e:
            # Note that this is flying out of the unprotection (ie. the
            # security context), which is trusted to not leak unintended
            # information in unencrypted responses. (By comparison, a
            # renderable exception flying out of a user
            # render_to_pipe could only be be rendered to a
            # protected message, and we'd need to be weary of rendering errors
            # during to_message as well).
            #
            # Note that this clause is not a no-op: it protects the 4.01 Echo
            # recovery exception (which is also a ReplayError) from being
            # treated as such.
            raise e
        # The other errors could be ported thee but would need some better NoResponse handling.
        except oscore.ReplayError:
            if request.mtype == aiocoap.CON:
                pipe.add_response(
                    aiocoap.Message(
                        code=aiocoap.UNAUTHORIZED, max_age=0, payload=b"Replay detected"
                    ),
                    is_last=True,
                )
            return
        except oscore.DecodeError:
            if request.mtype == aiocoap.CON:
                raise error.BadOption("Failed to decode COSE")
            else:
                return
        except oscore.ProtectionInvalid:
            if request.mtype == aiocoap.CON:
                raise error.BadRequest("Decryption failed")
            else:
                return

        unprotected.remote = OSCOREAddress(sc, request.remote)

        self.log.debug("Request %r was unprotected into %r", request, unprotected)

        sc = sc.context_for_response()

        inner_pipe = aiocoap.pipe.IterablePipe(unprotected)
        pr_that_can_take_errors = aiocoap.pipe.error_to_message(inner_pipe, self.log)
        # FIXME: do not create a task but run this in here (can this become a
        # feature of the aiterable PR?)
        aiocoap.pipe.run_driving_pipe(
            pr_that_can_take_errors,
            self._inner_site.render_to_pipe(inner_pipe),
            name="OSCORE response rendering for %r" % unprotected,
        )

        async for event in inner_pipe:
            if event.exception is not None:
                # These are expected to be rare in handlers
                #
                # FIXME should we try to render them? (See also
                # run_driving_pipe). Just raising them
                # would definitely be bad, as they might be renderable and
                # then would hit the outer message.
                self.log.warn(
                    "Turning error raised from renderer into nondescript protected error %r",
                    event.exception,
                )
                message = aiocoap.Message(code=aiocoap.INTERNAL_SERVER_ERROR)
                is_last = True
            else:
                message = event.message
                is_last = event.is_last

            # FIXME: Around several places in the use of pipe (and
            # now even here), non-final events are hard-coded as observations.
            # This should shift toward the source telling, or the stream being
            # annotated as "eventually consistent resource states".
            if not is_last:
                message.opt.observe = 0

            protected_response, _ = sc.protect(message, seqno)
            if message.opt.observe is not None:
                # FIXME: should be done in protect, or by something else that
                # generally handles obs numbers better (sending the
                # oscore-reconstructed number is nice because it's consistent
                # with a proxy that doesn't want to keep a counter when it
                # knows it's OSCORE already), but starting this per obs with
                # zero (unless it was done on that token recently) would be
                # most efficient
                protected_response.opt.observe = sc.sender_sequence_number & 0xFFFFFFFF
            self.log.debug(
                "Response %r was encrypted into %r", message, protected_response
            )

            pipe.add_response(protected_response, is_last=is_last)
            if event.is_last:
                break
        # The created task gets cancelled here because the __aiter__ result is
        # dropped and thus all interest in the inner_pipe goes away

    async def _render_edhoc_to_pipe(self, pipe):
        self.log.debug("Processing request as EDHOC message 1 or 3")
        # Conveniently we don't have to care for observation, and thus can treat the rendering to a pipeline as just a rendering

        request = pipe.request

        if request.code is not POST:
            raise error.MethodNotAllowed

        if any(
            o.number.is_critical()
            for o in request.opt.option_list()
            if o.number not in (OptionNumber.URI_PATH, OptionNumber.URI_HOST)
        ):
            # FIXME: This should be done by every resource handler (see
            # https://github.com/chrysn/aiocoap/issues/268) -- this is crude
            # but better than doing nothing (and because we're rendering to a
            # pipe, chances are upcoming mitigation might not catch this)
            raise error.BadOption

        if len(request.payload) == 0:
            raise error.BadRequest

        if request.payload[0:1] == cbor2.dumps(True):
            self.log.debug("Processing request as EDHOC message 1")
            self._process_edhoc_msg12(pipe)
        else:
            self.log.debug("Processing request as EDHOC message 3")
            self._process_edhoc_msg34(pipe)

    def _process_edhoc_msg12(self, pipe):
        request = pipe.request

        origin = request.get_request_uri(local_is_server=True).removesuffix(
            "/.well-known/edhoc"
        )
        own_credential_object = self._get_edhoc_identity(origin)
        if own_credential_object is None:
            self.log.error(
                "Peer attempted EDHOC even though no EDHOC credentials are configured for %s",
                origin,
            )
            raise error.NotFound

        # FIXME lakers: Shouldn't have to commit this early, might still look at EAD1
        assert isinstance(own_credential_object.own_cred, dict) and list(
            own_credential_object.own_cred.keys()
        ) == [14], (
            "So far can only process CCS style own credentials a la {14: ...}, own_cred = %r"
            % own_credential_object.own_cred
        )
        responder = lakers.EdhocResponder(
            r=own_credential_object.own_key.d,
            cred_r=cbor2.dumps(own_credential_object.own_cred[14], canonical=True),
        )
        c_i, ead_1 = responder.process_message_1(request.payload[1:])
        if ead_1 is not None:
            self.log.error("Aborting EDHOC: EAD1 present")
            raise error.BadRequest

        used_own_identifiers = (
            self.server_credentials.find_all_used_contextless_oscore_kid()
        )
        # can't have c_r==c_i
        used_own_identifiers.add(c_i)
        # FIXME try larger ones too, but currently they wouldn't work in Lakers
        candidates = [cbor2.dumps(i) for i in range(-24, 24)]
        candidates = [c for c in candidates if c not in used_own_identifiers]
        if not candidates:
            # FIXME: LRU or timeout the contexts
            raise error.InternalServerError("Too many contexts")
        c_r = candidates[0]
        message_2 = responder.prepare_message_2(
            own_credential_object.own_cred_style.as_lakers(), c_r, None
        )

        credentials_entry = edhoc.EdhocResponderContext(
            responder,
            c_i,
            c_r,
            self.server_credentials,
            self.log,
        )
        # FIXME we shouldn't need arbitrary keys
        self.server_credentials[":" + uuid.uuid4().hex] = credentials_entry

        pipe.add_response(
            aiocoap.Message(code=aiocoap.CHANGED, payload=message_2), is_last=True
        )

    def _process_edhoc_msg34(self, pipe):
        request = pipe.request

        payload = io.BytesIO(request.payload)
        try:
            c_r = cbor2.load(payload)
        except cbor2.CBORDecodeError:
            self.log.error("Message 3 received without valid CBOR start")
            raise error.BadRequest
        message_3 = payload.read()

        if isinstance(c_r, int) and not isinstance(c_r, bool) and -24 <= c_r < 23:
            c_r = cbor2.dumps(c_r)
        if not isinstance(c_r, bytes):
            self.log.error(f"Message 3 received with invalid C_R {c_r:r}")
            raise error.BadRequest

        # Our lookup is modelled expecting OSCORE header objects, so we rebuild one
        unprotected = {oscore.COSE_KID: c_r}

        try:
            sc = self.server_credentials.find_oscore(unprotected)
        except KeyError:
            self.log.error(
                f"No OSCORE context found with recipient_id / c_r matching {c_r!r}"
            )
            raise error.BadRequest

        if not isinstance(sc, edhoc.EdhocResponderContext):
            raise error.BadRequest

        message_4 = sc._offer_message_3(message_3)

        pipe.add_response(
            aiocoap.Message(code=aiocoap.CHANGED, payload=message_4), is_last=True
        )

    def _get_edhoc_identity(self, origin: str) -> Optional[edhoc.EdhocCredentials]:
        """With lakers-python 0.3.1, we can effectively only have one identity
        per host; expect this to change once we gain access to EAD1 (plus more
        when there are more methods or cipher suites)
        """

        # That this works is a flaw of the credentials format by itself
        candidate = self.server_credentials.get(origin + "/*")
        if not isinstance(candidate, edhoc.EdhocCredentials):
            # FIXME not really a pair needed is it?
            return None
        return candidate



#!/usr/bin/env python3
# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from pathlib import Path
import shutil
import urllib.request

out = Path(__file__).parent / "for-launcher"
out.mkdir(exist_ok=True)
contrib = Path(__file__).parent.parent

with (out / "android.txt").open("w") as androidtxt:
    androidtxt.write("title=aiocoap kivy widget\n")
    androidtxt.write("author=chrysn\n")
    androidtxt.write("orientation=portrait\n")

shutil.copytree(contrib / "widgets_common", out / "widgets_common", dirs_exist_ok=True)
shutil.copytree(contrib.parent / "aiocoap", out / "aiocoap", dirs_exist_ok=True)

wheels = [
    "https://files.pythonhosted.org/packages/d5/e1/af78b099196feaab7c0252108abc4f5cfd36d255ac47c4b4a695ff838bf9/cbor2-5.4.6-py3-none-any.whl",
    "https://files.pythonhosted.org/packages/93/69/e391bd51bc08ed9141ecd899a0ddb61ab6465309f1eb470905c0c8868081/docutils-0.19-py3-none-any.whl",
]

wheelfilenames = []

for w in wheels:
    with urllib.request.urlopen(w) as request:
        wheelname = w[w.rindex("/") + 1 :]
        with (out / wheelname).open("wb") as wheel_out:
            wheel_out.write(request.read())
        wheelfilenames.append(wheelname)

with (out / "main.py").open("w") as mainpy:
    mainpy.write("import sys\n")
    mainpy.write("from pathlib import Path\n")
    mainpy.write("_here = Path(__file__).parent\n")
    for w in wheelfilenames:
        mainpy.write(f"sys.path.append(str(_here / '{w}') + '/')\n")
    with (contrib / "aiocoap-kivy-widget").open() as infile:
        mainpy.write(infile.read())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This module contains the classes that are responsible for keeping track of
messages:

*   :class:`Context` roughly represents the CoAP endpoint (basically a UDP
    socket) -- something that can send requests and possibly can answer
    incoming requests.

    Incoming requests are processed in tasks created by the context.

*   a :class:`Request` gets generated whenever a request gets sent to keep
    track of the response

Logging
~~~~~~~

Several constructors of the Context accept a logger name; these names go into
the construction of a Python logger.

Log events will be emitted to these on different levels, with "warning" and
above being a practical default for things that should may warrant reviewing by
an operator:

* DEBUG is used for things that occur even under perfect conditions.
* INFO is for things that are well expected, but might be interesting during
  testing a network of nodes and not just when debugging the library. (This
  includes timeouts, retransmissions, and pings.)
* WARNING is for everything that indicates a malbehaved peer. These don't
  *necessarily* indicate a client bug, though: Things like requesting a
  nonexistent block can just as well happen when a resource's content has
  changed between blocks. The library will not go out of its way to determine
  whether there is a plausible explanation for the odd behavior, and will
  report something as a warning in case of doubt.
* ERROR is used when something clearly went wrong. This includes irregular
  connection terminations and resource handler errors (which are demoted to
  error responses), and can often contain a backtrace.

Logs will generally reveal messages exchanged between this and other systems,
and attackers can observe their encrypted counterparts. Private or shared keys
are only logged through an internal `log_secret` function, which usually
replaces them with a redacted value. Setting the ``AIOCOAP_REVEAL_KEYS``
environment variable to the value ``show secrets in logs`` bypasses that
mechanism. As an additional precaution, this is only accepted if the effective
user has write access to the aiocoap source code.
"""

import asyncio
import weakref
import time
from typing import Optional, List

from . import defaults
from .credentials import CredentialsMap
from .message import Message
from .messagemanager import MessageManager
from .tokenmanager import TokenManager
from .pipe import Pipe, run_driving_pipe, error_to_message
from . import interfaces
from . import error
from .numbers import INTERNAL_SERVER_ERROR, NOT_FOUND, CONTINUE, SHUTDOWN_TIMEOUT

import warnings
import logging


class Context(interfaces.RequestProvider):
    """Applications' entry point to the network

    A :class:`.Context` coordinates one or more network :mod:`.transports`
    implementations and dispatches data between them and the application.

    The application can start requests using the message dispatch methods, and
    set a :class:`resources.Site` that will answer requests directed to the
    application as a server.

    On the library-internals side, it is the prime implementation of the
    :class:`interfaces.RequestProvider` interface, creates :class:`Request` and
    :class:`Response` classes on demand, and decides which transport
    implementations to start and which are to handle which messages.

    **Context creation and destruction**

    The following functions are provided for creating and stopping a context:

    .. note::

        A typical application should only ever create one context, even (or
        especially when) it acts both as a server and as a client (in which
        case a server context should be created).

        A context that is not used any more must be shut down using
        :meth:`.shutdown()`, but typical applications will not need to because
        they use the context for the full process lifetime.

    .. automethod:: create_client_context
    .. automethod:: create_server_context

    .. automethod:: shutdown

    **Dispatching messages**

    CoAP requests can be sent using the following functions:

    .. automethod:: request

    If more control is needed, you can create a :class:`Request` yourself and
    pass the context to it.


    **Other methods and properties**

    The remaining methods and properties are to be considered unstable even
    when the project reaches a stable version number; please file a feature
    request for stabilization if you want to reliably access any of them.
    """

    def __init__(
        self,
        loop=None,
        serversite=None,
        loggername="coap",
        client_credentials=None,
        server_credentials=None,
    ):
        self.log = logging.getLogger(loggername)

        self.loop = loop or asyncio.get_event_loop()

        self.serversite = serversite

        self.request_interfaces = []

        self.client_credentials = client_credentials or CredentialsMap()
        self.server_credentials = server_credentials or CredentialsMap()

    #
    # convenience methods for class instanciation
    #

    async def _append_tokenmanaged_messagemanaged_transport(
        self, message_interface_constructor
    ):
        tman = TokenManager(self)
        mman = MessageManager(tman)
        transport = await message_interface_constructor(mman)

        mman.message_interface = transport
        tman.token_interface = mman

        self.request_interfaces.append(tman)

    async def _append_tokenmanaged_transport(self, token_interface_constructor):
        tman = TokenManager(self)
        transport = await token_interface_constructor(tman)

        tman.token_interface = transport

        self.request_interfaces.append(tman)

    @classmethod
    async def create_client_context(
        cls, *, loggername="coap", loop=None, transports: Optional[List[str]] = None
    ):
        """Create a context bound to all addresses on a random listening port.

        This is the easiest way to get a context suitable for sending client
        requests.

        :meta private:
            (not actually private, just hiding from automodule due to being
            grouped with the important functions)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        self = cls(loop=loop, serversite=None, loggername=loggername)

        selected_transports = transports or defaults.get_default_clienttransports(
            loop=loop
        )

        # FIXME make defaults overridable (postponed until they become configurable too)
        for transportname in selected_transports:
            if transportname == "udp6":
                from .transports.udp6 import MessageInterfaceUDP6

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceUDP6.create_client_transport_endpoint(
                        mman, log=self.log, loop=loop
                    )
                )
            elif transportname == "simple6":
                from .transports.simple6 import MessageInterfaceSimple6

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimple6.create_client_transport_endpoint(
                        mman, log=self.log, loop=loop
                    )
                )
            elif transportname == "tinydtls":
                from .transports.tinydtls import MessageInterfaceTinyDTLS

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceTinyDTLS.create_client_transport_endpoint(
                        mman, log=self.log, loop=loop
                    )
                )
            elif transportname == "tcpclient":
                from .transports.tcp import TCPClient

                await self._append_tokenmanaged_transport(
                    lambda tman: TCPClient.create_client_transport(tman, self.log, loop)
                )
            elif transportname == "tlsclient":
                from .transports.tls import TLSClient

                await self._append_tokenmanaged_transport(
                    lambda tman: TLSClient.create_client_transport(
                        tman, self.log, loop, self.client_credentials
                    )
                )
            elif transportname == "ws":
                from .transports.ws import WSPool

                await self._append_tokenmanaged_transport(
                    lambda tman: WSPool.create_transport(
                        tman, self.log, loop, client_credentials=self.client_credentials
                    )
                )
            elif transportname == "oscore":
                from .transports.oscore import TransportOSCORE

                oscoretransport = TransportOSCORE(self, self)
                self.request_interfaces.append(oscoretransport)
            else:
                raise RuntimeError(
                    "Transport %r not know for client context creation" % transportname
                )

        return self

    @classmethod
    async def create_server_context(
        cls,
        site,
        bind=None,
        *,
        loggername="coap-server",
        loop=None,
        _ssl_context=None,
        multicast=[],
        server_credentials=None,
        transports: Optional[List[str]] = None,
    ):
        """Create a context, bound to all addresses on the CoAP port (unless
        otherwise specified in the ``bind`` argument).

        This is the easiest way to get a context suitable both for sending
        client and accepting server requests.

        The ``bind`` argument, if given, needs to be a 2-tuple of IP address
        string and port number, where the port number can be None to use the default port.

        If ``multicast`` is given, it needs to be a list of (multicast address,
        interface name) tuples, which will all be joined. (The IPv4 style of
        selecting the interface by a local address is not supported; users may
        want to use the netifaces package to arrive at an interface name for an
        address).

        As a shortcut, the list may also contain interface names alone. Those
        will be joined for the 'all CoAP nodes' groups of IPv4 and IPv6 (with
        scopes 2 and 5) as well as the respective 'all nodes' groups in IPv6.

        Under some circumstances you may already need a context to pass into
        the site for creation; this is typically the case for servers that
        trigger requests on their own. For those cases, it is usually easiest
        to pass None in as a site, and set the fully constructed site later by
        assigning to the ``serversite`` attribute.

        :meta private:
            (not actually private, just hiding from automodule due to being
            grouped with the important functions)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        self = cls(
            loop=loop,
            serversite=site,
            loggername=loggername,
            server_credentials=server_credentials,
        )

        multicast_done = not multicast

        selected_transports = transports or defaults.get_default_servertransports(
            loop=loop
        )

        for transportname in selected_transports:
            if transportname == "udp6":
                from .transports.udp6 import MessageInterfaceUDP6

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceUDP6.create_server_transport_endpoint(
                        mman, log=self.log, loop=loop, bind=bind, multicast=multicast
                    )
                )
                multicast_done = True
            # FIXME this is duplicated from the client version, as those are client-only anyway
            elif transportname == "simple6":
                from .transports.simple6 import MessageInterfaceSimple6

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimple6.create_client_transport_endpoint(
                        mman, log=self.log, loop=loop
                    )
                )
            elif transportname == "tinydtls":
                from .transports.tinydtls import MessageInterfaceTinyDTLS

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceTinyDTLS.create_client_transport_endpoint(
                        mman, log=self.log, loop=loop
                    )
                )
            # FIXME end duplication
            elif transportname == "tinydtls_server":
                from .transports.tinydtls_server import MessageInterfaceTinyDTLSServer

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceTinyDTLSServer.create_server(
                        bind,
                        mman,
                        log=self.log,
                        loop=loop,
                        server_credentials=self.server_credentials,
                    )
                )
            elif transportname == "simplesocketserver":
                from .transports.simplesocketserver import MessageInterfaceSimpleServer

                await self._append_tokenmanaged_messagemanaged_transport(
                    lambda mman: MessageInterfaceSimpleServer.create_server(
                        bind, mman, log=self.log, loop=loop
                    )
                )
            elif transportname == "tcpserver":
                from .transports.tcp import TCPServer

                await self._append_tokenmanaged_transport(
                    lambda tman: TCPServer.create_server(bind, tman, self.log, loop)
                )
            elif transportname == "tcpclient":
                from .transports.tcp import TCPClient

                await self._append_tokenmanaged_transport(
                    lambda tman: TCPClient.create_client_transport(tman, self.log, loop)
                )
            elif transportname == "tlsserver":
                if _ssl_context is not None:
                    from .transports.tls import TLSServer

                    await self._append_tokenmanaged_transport(
                        lambda tman: TLSServer.create_server(
                            bind, tman, self.log, loop, _ssl_context
                        )
                    )
            elif transportname == "tlsclient":
                from .transports.tls import TLSClient

                await self._append_tokenmanaged_transport(
                    lambda tman: TLSClient.create_client_transport(
                        tman, self.log, loop, self.client_credentials
                    )
                )
            elif transportname == "ws":
                from .transports.ws import WSPool

                await self._append_tokenmanaged_transport(
                    # None, None: Unlike the other transports this has a server/client generic creator, and only binds if there is some bind
                    lambda tman: WSPool.create_transport(
                        tman,
                        self.log,
                        loop,
                        client_credentials=self.client_credentials,
                        server_bind=bind or (None, None),
                        server_context=_ssl_context,
                    )
                )
            elif transportname == "oscore":
                from .transports.oscore import TransportOSCORE

                oscoretransport = TransportOSCORE(self, self)
                self.request_interfaces.append(oscoretransport)
            else:
                raise RuntimeError(
                    "Transport %r not know for server context creation" % transportname
                )

        if not multicast_done:
            self.log.warning(
                "Multicast was requested, but no multicast capable transport was selected."
            )

        # This is used in tests to wait for externally launched servers to be ready
        self.log.debug("Server ready to receive requests")

        return self

    async def shutdown(self):
        """Take down any listening sockets and stop all related timers.

        After this coroutine terminates, and once all external references to
        the object are dropped, it should be garbage-collectable.

        This method takes up to
        :const:`aiocoap.numbers.constants.SHUTDOWN_TIMEOUT` seconds, allowing
        transports to perform any cleanup implemented in them (such as orderly
        connection shutdown and cancelling observations, where the latter is
        currently not implemented).

        :meta private:
            (not actually private, just hiding from automodule due to being
            grouped with the important functions)
        """

        self.log.debug("Shutting down context")

        done, pending = await asyncio.wait(
            [
                asyncio.create_task(
                    ri.shutdown(),
                    name="Shutdown of %r" % ri,
                )
                for ri in self.request_interfaces
            ],
            timeout=SHUTDOWN_TIMEOUT,
        )
        for item in done:
            await item
        if pending:
            # Apart from being useful to see, this also ensures that developers
            # see the error in the logs during test suite runs -- and the error
            # should be easier to follow than the "we didn't garbage collect
            # everything" errors we see anyway (or otherwise, if the error is
            # escalated into a test failure)
            self.log.error(
                "Shutdown timeout exceeded, returning anyway. Interfaces still busy: %s",
                pending,
            )

    # FIXME: determine how official this should be, or which part of it is
    # public -- now that BlockwiseRequest uses it. (And formalize what can
    # change about messages and what can't after the remote has been thusly
    # populated).
    async def find_remote_and_interface(self, message):
        if message.remote is None:
            raise error.MissingRemoteError()
        for ri in self.request_interfaces:
            if await ri.fill_or_recognize_remote(message):
                return ri
        raise error.NoRequestInterface()

    def request(self, request_message, handle_blockwise=True):
        if handle_blockwise:
            return BlockwiseRequest(self, request_message)

        pipe = Pipe(request_message, self.log)
        # Request sets up callbacks at creation
        result = Request(pipe, self.loop, self.log)

        async def send():
            try:
                request_interface = await self.find_remote_and_interface(
                    request_message
                )
                request_interface.request(pipe)
            except Exception as e:
                pipe.add_exception(e)
                return

        self.loop.create_task(
            send(),
            name="Request processing of %r" % result,
        )
        return result

    # the following are under consideration for moving into Site or something
    # mixed into it

    def render_to_pipe(self, pipe):
        """Fill a pipe by running the site's render_to_pipe interface and
        handling errors."""

        pr_that_can_receive_errors = error_to_message(pipe, self.log)

        run_driving_pipe(
            pr_that_can_receive_errors,
            self._render_to_pipe(pipe),
            name="Rendering for %r" % pipe.request,
        )

    async def _render_to_pipe(self, pipe):
        if self.serversite is None:
            pipe.add_response(
                Message(code=NOT_FOUND, payload=b"not a server"), is_last=True
            )
            return

        return await self.serversite.render_to_pipe(pipe)


class BaseRequest(object):
    """Common mechanisms of :class:`Request` and :class:`MulticastRequest`"""


class BaseUnicastRequest(BaseRequest):
    """A utility class that offers the :attr:`response_raising` and
    :attr:`response_nonraising` alternatives to waiting for the
    :attr:`response` future whose error states can be presented either as an
    unsuccessful response (eg. 4.04) or an exception.

    It also provides some internal tools for handling anything that has a
    :attr:`response` future and an :attr:`observation`"""

    @property
    async def response_raising(self):
        """An awaitable that returns if a response comes in and is successful,
        otherwise raises generic network exception or a
        :class:`.error.ResponseWrappingError` for unsuccessful responses.

        Experimental Interface."""

        response = await self.response
        if not response.code.is_successful():
            raise error.ResponseWrappingError(response)

        return response

    @property
    async def response_nonraising(self):
        """An awaitable that rather returns a 500ish fabricated message (as a
        proxy would return) instead of raising an exception.

        Experimental Interface."""

        # FIXME: Can we smuggle error_to_message into the underlying pipe?
        # That should make observe notifications into messages rather
        # than exceptions as well, plus it has fallbacks for `e.to_message()`
        # raising.

        try:
            return await self.response
        except error.RenderableError as e:
            return e.to_message()
        except Exception:
            return Message(code=INTERNAL_SERVER_ERROR)


class Request(interfaces.Request, BaseUnicastRequest):
    # FIXME: Implement timing out with REQUEST_TIMEOUT here

    def __init__(self, pipe, loop, log):
        self._pipe = pipe

        self.response = loop.create_future()

        if pipe.request.opt.observe == 0:
            self.observation = ClientObservation()
        else:
            self.observation = None

        self._runner = self._run()
        self._runner.send(None)

        def process(event):
            try:
                # would be great to have self or the runner as weak ref, but
                # see ClientObservation.register_callback comments -- while
                # that is around, we can't weakref here.
                self._runner.send(event)
                return True
            except StopIteration:
                return False

        self._stop_interest = self._pipe.on_event(process)

        self.log = log

        self.response.add_done_callback(self._response_cancellation_handler)

    def _response_cancellation_handler(self, response):
        # Propagate cancellation to the runner (if interest in the first
        # response is lost, there won't be observation items to pull out), but
        # not general completion (because if it's completed and not cancelled,
        # eg. when an observation is active)
        if self.response.cancelled() and self._runner is not None:
            # Dropping the only reference makes it stop with GeneratorExit,
            # similar to a cancelled task
            self._runner = None
            self._stop_interest()
        # Otherwise, there will be a runner still around, and it's its task to
        # call _stop_interest.

    @staticmethod
    def _add_response_properties(response, request):
        response.request = request

    def _run(self):
        # FIXME: This is in iterator form because it used to be a task that
        # awaited futures, and that code could be easily converted to an
        # iterator. I'm not sure that's a bad state here, but at least it
        # should be a more conscious decision to make this an iterator rather
        # than just having it happen to be one.
        #
        # FIXME: check that responses come from the same remmote as long as we're assuming unicast

        first_event = yield None

        if first_event.message is not None:
            self._add_response_properties(first_event.message, self._pipe.request)
            self.response.set_result(first_event.message)
        else:
            self.response.set_exception(first_event.exception)
            if not isinstance(first_event.exception, error.Error):
                self.log.warning(
                    "An exception that is not an aiocoap Error was raised "
                    "from a transport; please report this as a bug in "
                    "aiocoap: %r",
                    first_event.exception,
                )

        if self.observation is None:
            if not first_event.is_last:
                self.log.error(
                    "Pipe indicated more possible responses"
                    " while the Request handler would not know what to"
                    " do with them, stopping any further request."
                )
                self._stop_interest()
            return

        if first_event.is_last:
            self.observation.error(error.NotObservable())
            return

        if first_event.message.opt.observe is None:
            self.log.error(
                "Pipe indicated more possible responses"
                " while the Request handler would not know what to"
                " do with them, stopping any further request."
            )
            self._stop_interest()
            return

        # variable names from RFC7641 Section 3.4
        v1 = first_event.message.opt.observe
        t1 = time.time()

        while True:
            # We don't really support cancellation of observations yet (see
            # https://github.com/chrysn/aiocoap/issues/92), but at least
            # stopping the interest is a way to free the local resources after
            # the first observation update, and to make the MID handler RST the
            # observation on the next.
            # FIXME: there *is* now a .on_cancel callback, we should at least
            # hook into that, and possibly even send a proper cancellation
            # then.
            next_event = yield True
            if self.observation.cancelled:
                self._stop_interest()
                return

            if next_event.exception is not None:
                self.observation.error(next_event.exception)
                if not next_event.is_last:
                    self._stop_interest()
                if not isinstance(next_event.exception, error.Error):
                    self.log.warning(
                        "An exception that is not an aiocoap Error was "
                        "raised from a transport during an observation; "
                        "please report this as a bug in aiocoap: %r",
                        next_event.exception,
                    )
                return

            self._add_response_properties(next_event.message, self._pipe.request)

            if next_event.message.opt.observe is not None:
                # check for reordering
                v2 = next_event.message.opt.observe
                t2 = time.time()

                is_recent = (
                    (v1 < v2 and v2 - v1 < 2**23)
                    or (v1 > v2 and v1 - v2 > 2**23)
                    or (
                        t2
                        > t1
                        + self._pipe.request.transport_tuning.OBSERVATION_RESET_TIME
                    )
                )
                if is_recent:
                    t1 = t2
                    v1 = v2
            else:
                # the terminal message is always the last
                is_recent = True

            if is_recent:
                self.observation.callback(next_event.message)

            if next_event.is_last:
                self.observation.error(error.ObservationCancelled())
                return

            if next_event.message.opt.observe is None:
                self.observation.error(error.ObservationCancelled())
                self.log.error(
                    "Pipe indicated more possible responses"
                    " while the Request handler would not know what to"
                    " do with them, stopping any further request."
                )
                self._stop_interest()
                return


class BlockwiseRequest(BaseUnicastRequest, interfaces.Request):
    def __init__(self, protocol, app_request):
        self.protocol = protocol
        self.log = self.protocol.log.getChild("blockwise-requester")

        self.response = protocol.loop.create_future()

        if app_request.opt.observe is not None:
            self.observation = ClientObservation()
        else:
            self.observation = None

        self._runner = protocol.loop.create_task(
            self._run_outer(
                app_request,
                self.response,
                weakref.ref(self.observation)
                if self.observation is not None
                else lambda: None,
                self.protocol,
                self.log,
            ),
            name="Blockwise runner for %r" % app_request,
        )
        self.response.add_done_callback(self._response_cancellation_handler)

    def _response_cancellation_handler(self, response_future):
        # see Request._response_cancellation_handler
        if self.response.cancelled():
            self._runner.cancel()

    @classmethod
    async def _run_outer(cls, app_request, response, weak_observation, protocol, log):
        try:
            await cls._run(app_request, response, weak_observation, protocol, log)
        except asyncio.CancelledError:
            pass  # results already set
        except Exception as e:
            logged = False
            if not response.done():
                logged = True
                response.set_exception(e)
            obs = weak_observation()
            if app_request.opt.observe is not None and obs is not None:
                logged = True
                obs.error(e)
            if not logged:
                # should be unreachable
                log.error(
                    "Exception in BlockwiseRequest runner neither went to response nor to observation: %s",
                    e,
                    exc_info=e,
                )

    # This is a class method because that allows self and self.observation to
    # be freed even when this task is running, and the task to stop itself --
    # otherwise we couldn't know when users just "forget" about a request
    # object after using its response (esp. in observe cases) and leave this
    # task running.
    @classmethod
    async def _run(cls, app_request, response, weak_observation, protocol, log):
        # we need to populate the remote right away, because the choice of
        # blocks depends on it.
        await protocol.find_remote_and_interface(app_request)

        size_exp = app_request.remote.maximum_block_size_exp

        if app_request.opt.block1 is not None:
            warnings.warn(
                "Setting a block1 option in a managed block-wise transfer is deprecated. Instead, set request.remote.maximum_block_size_exp to the desired value",
                DeprecationWarning,
                stacklevel=2,
            )
            assert app_request.opt.block1.block_number == 0, (
                "Unexpected block number in app_request"
            )
            assert not app_request.opt.block1.more, (
                "Unexpected more-flag in app_request"
            )
            # this is where the library user can traditionally pass in size
            # exponent hints into the library.
            size_exp = app_request.opt.block1.size_exponent

        # Offset in the message in blocks of size_exp. Whoever changes size_exp
        # is responsible for updating this number.
        block_cursor = 0

        while True:
            # ... send a chunk

            if size_exp >= 6:
                # FIXME from maximum_payload_size
                fragmentation_threshold = app_request.remote.maximum_payload_size
            else:
                fragmentation_threshold = 2 ** (size_exp + 4)

            if (
                app_request.opt.block1 is not None
                or len(app_request.payload) > fragmentation_threshold
            ):
                current_block1 = app_request._extract_block(
                    block_cursor, size_exp, app_request.remote.maximum_payload_size
                )
                if block_cursor == 0:
                    current_block1.opt.size1 = len(app_request.payload)
            else:
                current_block1 = app_request

            blockrequest = protocol.request(current_block1, handle_blockwise=False)
            blockresponse = await blockrequest.response

            # store for future blocks to ensure that the next blocks will be
            # sent from the same source address (in the UDP case; for many
            # other transports it won't matter). carrying along locally set block size limitation
            if (
                app_request.remote.maximum_block_size_exp
                < blockresponse.remote.maximum_block_size_exp
            ):
                blockresponse.remote.maximum_block_size_exp = (
                    app_request.remote.maximum_block_size_exp
                )
            app_request.remote = blockresponse.remote

            if blockresponse.opt.block1 is None:
                if blockresponse.code.is_successful() and current_block1.opt.block1:
                    log.warning(
                        "Block1 option completely ignored by server, assuming it knows what it is doing."
                    )
                # FIXME: handle 4.13 and retry with the indicated size option
                break

            block1 = blockresponse.opt.block1
            log.debug(
                "Response with Block1 option received, number = %d, more = %d, size_exp = %d.",
                block1.block_number,
                block1.more,
                block1.size_exponent,
            )

            if block1.block_number != current_block1.opt.block1.block_number:
                raise error.UnexpectedBlock1Option("Block number mismatch")

            if size_exp == 7:
                block_cursor += len(current_block1.payload) // 1024
            else:
                block_cursor += 1

            while block1.size_exponent < size_exp:
                block_cursor *= 2
                size_exp -= 1

            if not current_block1.opt.block1.more:
                if block1.more or blockresponse.code == CONTINUE:
                    # treating this as a protocol error -- letting it slip
                    # through would misrepresent the whole operation as an
                    # over-all 2.xx (successful) one.
                    raise error.UnexpectedBlock1Option(
                        "Server asked for more data at end of body"
                    )
                break

            # checks before preparing the next round:

            if blockresponse.opt.observe:
                # we're not *really* interested in that block, we just sent an
                # observe option to indicate that we'll want to observe the
                # resulting representation as a whole
                log.warning(
                    "Server answered Observe in early Block1 phase, cancelling the erroneous observation."
                )
                blockrequest.observe.cancel()

            if block1.more:
                # FIXME i think my own server is dowing this wrong
                # if response.code != CONTINUE:
                #    raise error.UnexpectedBlock1Option("more-flag set but no Continue")
                pass
            else:
                if not blockresponse.code.is_successful():
                    break
                else:
                    # ignoring (discarding) the successul intermediate result, waiting for a final one
                    continue

        lower_observation = None
        if app_request.opt.observe is not None:
            if blockresponse.opt.observe is not None:
                lower_observation = blockrequest.observation
            else:
                obs = weak_observation()
                if obs:
                    obs.error(error.NotObservable())
                del obs

        assert blockresponse is not None, "Block1 loop broke without setting a response"
        blockresponse.opt.block1 = None

        # FIXME check with RFC7959: it just says "send requests similar to the
        # requests in the Block1 phase", what does that mean? using the last
        # block1 as a reference for now, especially because in the
        # only-one-request-block case, that's the original request we must send
        # again and again anyway
        assembled_response = await cls._complete_by_requesting_block2(
            protocol, current_block1, blockresponse, log
        )

        response.set_result(assembled_response)
        # finally set the result

        if lower_observation is not None:
            # FIXME this can all be simplified a lot since it's no more
            # expected that observations shut themselves down when GC'd.
            obs = weak_observation()
            del weak_observation
            if obs is None:
                lower_observation.cancel()
                return
            future_weak_observation = protocol.loop.create_future()  # packing this up because its destroy callback needs to reference the subtask
            subtask = asyncio.create_task(
                cls._run_observation(
                    app_request,
                    lower_observation,
                    future_weak_observation,
                    protocol,
                    log,
                ),
                name="Blockwise observation for %r" % app_request,
            )
            future_weak_observation.set_result(
                weakref.ref(obs, lambda obs: subtask.cancel())
            )
            obs.on_cancel(subtask.cancel)
            del obs
            await subtask

    @classmethod
    async def _run_observation(
        cls, original_request, lower_observation, future_weak_observation, protocol, log
    ):
        weak_observation = await future_weak_observation
        # we can use weak_observation() here at any time, because whenever that
        # becomes None, this task gets cancelled
        try:
            async for block1_notification in lower_observation:
                log.debug("Notification received")
                full_notification = await cls._complete_by_requesting_block2(
                    protocol, original_request, block1_notification, log
                )
                log.debug("Reporting completed notification")
                weak_observation().callback(full_notification)
            # FIXME verify that this loop actually ends iff the observation
            # was cancelled -- otherwise find out the cause(s) or make it not
            # cancel under indistinguishable circumstances
            weak_observation().error(error.ObservationCancelled())
        except asyncio.CancelledError:
            return
        except Exception as e:
            weak_observation().error(e)
        finally:
            # We generally avoid idempotent cancellation, but we may have
            # reached this point either due to an earlier cancellation or
            # without one
            if not lower_observation.cancelled:
                lower_observation.cancel()

    @classmethod
    async def _complete_by_requesting_block2(
        cls, protocol, request_to_repeat, initial_response, log
    ):
        # FIXME this can probably be deduplicated against BlockwiseRequest

        if (
            initial_response.opt.block2 is None
            or initial_response.opt.block2.more is False
        ):
            initial_response.opt.block2 = None
            return initial_response

        if initial_response.opt.block2.block_number != 0:
            log.error("Error assembling blockwise response (expected first block)")
            raise error.UnexpectedBlock2()

        assembled_response = initial_response
        last_response = initial_response
        while True:
            current_block2 = request_to_repeat._generate_next_block2_request(
                assembled_response
            )

            current_block2 = current_block2.copy(remote=initial_response.remote)

            blockrequest = protocol.request(current_block2, handle_blockwise=False)
            last_response = await blockrequest.response

            if last_response.opt.block2 is None:
                log.warning(
                    "Server sent non-blockwise response after having started a blockwise transfer. Blockwise transfer cancelled, accepting single response."
                )
                return last_response

            block2 = last_response.opt.block2
            log.debug(
                "Response with Block2 option received, number = %d, more = %d, size_exp = %d.",
                block2.block_number,
                block2.more,
                block2.size_exponent,
            )
            try:
                assembled_response._append_response_block(last_response)
            except error.Error as e:
                log.error("Error assembling blockwise response, passing on error %r", e)
                raise

            if block2.more is False:
                return assembled_response


class ClientObservation:
    """An interface to observe notification updates arriving on a request.

    This class does not actually provide any of the observe functionality, it
    is purely a container for dispatching the messages via asynchronous
    iteration. It gets driven (ie. populated with responses or errors including
    observation termination) by a Request object.
    """

    def __init__(self):
        self.callbacks = []
        self.errbacks = []

        self.cancelled = False
        self._on_cancel = []

        self._latest_response = None
        # the analogous error is stored in _cancellation_reason when cancelled.

    def __aiter__(self):
        """`async for` interface to observations.

        This is the preferred interface to obtaining observations."""
        it = self._Iterator()
        self.register_callback(it.push, _suppress_deprecation=True)
        self.register_errback(it.push_err, _suppress_deprecation=True)
        return it

    class _Iterator:
        def __init__(self):
            self._future = asyncio.get_event_loop().create_future()

        def push(self, item):
            if self._future.done():
                # we don't care whether we overwrite anything, this is a lossy queue as observe is lossy
                self._future = asyncio.get_event_loop().create_future()
            self._future.set_result(item)

        def push_err(self, e):
            if self._future.done():
                self._future = asyncio.get_event_loop().create_future()
            self._future.set_exception(e)

        async def __anext__(self):
            f = self._future
            try:
                result = await self._future
                # FIXME see `await servobs._trigger` comment: might waiting for
                # the original future not yield the first future's result when
                # a quick second future comes in in a push?
                if f is self._future:
                    self._future = asyncio.get_event_loop().create_future()
                return result
            except (error.NotObservable, error.ObservationCancelled):
                # only exit cleanly when the server -- right away or later --
                # states that the resource is not observable any more
                # FIXME: check whether an unsuccessful message is still passed
                # as an observation result (or whether it should be)
                raise StopAsyncIteration

        def __del__(self):
            if self._future.done():
                try:
                    # Fetch the result so any errors show up at least in the
                    # finalizer output
                    self._future.result()
                except (error.ObservationCancelled, error.NotObservable):
                    # This is the case at the end of an observation cancelled
                    # by the server.
                    pass
                except error.NetworkError:
                    # This will already have shown up in the main result too.
                    pass
                except (error.LibraryShutdown, asyncio.CancelledError):
                    pass
                # Anything else flying out of this is unexpected and probably a
                # library error

    # When this function is removed, we can finally do cleanup better. Right
    # now, someone could register a callback that doesn't hold any references,
    # so we can't just stop the request when nobody holds a reference to this
    # any more. Once we're all in pull mode, we can make the `process` function
    # that sends data in here use a weak reference (because any possible
    # recipient would need to hold a reference to self or the iterator, and
    # thus _run).
    def register_callback(self, callback, _suppress_deprecation=False):
        """Call the callback whenever a response to the message comes in, and
        pass the response to it.

        The use of this function is deprecated: Use the asynchronous iteration
        interface instead."""
        if not _suppress_deprecation:
            warnings.warn(
                "register_callback on observe results is deprected: Use `async for notify in request.observation` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        if self.cancelled:
            return

        self.callbacks.append(callback)
        if self._latest_response is not None:
            callback(self._latest_response)

    def register_errback(self, callback, _suppress_deprecation=False):
        """Call the callback whenever something goes wrong with the
        observation, and pass an exception to the callback. After such a
        callback is called, no more callbacks will be issued.

        The use of this function is deprecated: Use the asynchronous iteration
        interface instead."""
        if not _suppress_deprecation:
            warnings.warn(
                "register_errback on observe results is deprected: Use `async for notify in request.observation` instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        if self.cancelled:
            callback(self._cancellation_reason)
            return
        self.errbacks.append(callback)

    def callback(self, response):
        """Notify all listeners of an incoming response"""

        self._latest_response = response

        for c in self.callbacks:
            c(response)

    def error(self, exception):
        """Notify registered listeners that the observation went wrong. This
        can only be called once."""

        if self.errbacks is None:
            raise RuntimeError(
                "Error raised in an already cancelled ClientObservation"
            ) from exception
        for c in self.errbacks:
            c(exception)

        self.cancel()
        self._cancellation_reason = exception

    def cancel(self):
        # FIXME determine whether this is called by anything other than error,
        # and make it private so there is always a _cancellation_reason
        """Cease to generate observation or error events. This will not
        generate an error by itself.

        This function is only needed while register_callback and
        register_errback are around; once their deprecations are acted on,
        dropping the asynchronous iterator will automatically cancel the
        observation.
        """

        assert not self.cancelled, "ClientObservation cancelled twice"

        # make sure things go wrong when someone tries to continue this
        self.errbacks = None
        self.callbacks = None

        self.cancelled = True
        while self._on_cancel:
            self._on_cancel.pop()()

        self._cancellation_reason = None

    def on_cancel(self, callback):
        if self.cancelled:
            callback()
        self._on_cancel.append(callback)

    def __repr__(self):
        return "<%s %s at %#x>" % (
            type(self).__name__,
            "(cancelled)"
            if self.cancelled
            else "(%s call-, %s errback(s))"
            % (len(self.callbacks), len(self.errbacks)),
            id(self),
        )


class ServerObservation:
    def __init__(self):
        self._accepted = False
        self._trigger = asyncio.get_event_loop().create_future()
        # A deregistration is "early" if it happens before the response message
        # is actually sent; calling deregister() in that time (typically during
        # `render()`) will not send an unsuccessful response message but just
        # sent this flag which is set to None as soon as it is too late for an
        # early deregistration.
        # This mechanism is temporary until more of aiocoap behaves like
        # Pipe which does not suffer from this limitation.
        self._early_deregister = False
        self._late_deregister = False

    def accept(self, cancellation_callback):
        self._accepted = True
        self._cancellation_callback = cancellation_callback

    def deregister(self, reason=None):
        if self._early_deregister is False:
            self._early_deregister = True
            return

        warnings.warn(
            "Late use of ServerObservation.deregister() is"
            " deprecated, use .trigger with an unsuccessful value"
            " instead",
            DeprecationWarning,
        )
        self.trigger(
            Message(code=INTERNAL_SERVER_ERROR, payload=b"Resource became unobservable")
        )

    def trigger(self, response=None, *, is_last=False):
        """Send an updated response; if None is given, the observed resource's
        rendering will be invoked to produce one.

        `is_last` can be set to True to indicate that no more responses will be
        sent. Note that an unsuccessful response will be the last no matter
        what is_last says, as such a message always terminates a CoAP
        observation."""
        if is_last:
            self._late_deregister = True
        if self._trigger.done():
            # we don't care whether we overwrite anything, this is a lossy queue as observe is lossy
            self._trigger = asyncio.get_event_loop().create_future()
        self._trigger.set_result(response)



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT
import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)


async def main():
    protocol = await Context.create_client_context()

    import aiocoap.oscore

    # Acting as Rikard Test 2 Entity 3
    protocol.client_credentials["coap://localhost/*"] = (
        aiocoap.oscore.SimpleGroupContext(
            algorithm=aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
            hashfun=aiocoap.oscore.hashfunctions[aiocoap.oscore.DEFAULT_HASHFUNCTION],
            alg_signature=aiocoap.oscore.Ed25519(),
            alg_group_enc=aiocoap.oscore.algorithms[aiocoap.oscore.DEFAULT_ALGORITHM],
            alg_pairwise_key_agreement=aiocoap.oscore.EcdhSsHkdf256(),
            group_id=bytes.fromhex("DD11"),
            master_secret=bytes.fromhex("11223344556677889900AABBCCDDEEFF"),
            master_salt=bytes.fromhex("1F2E3D4C5B6A7081"),
            # these are really moot given we only access this as a deterministic client
            sender_id=bytes.fromhex("52"),
            private_key=bytes.fromhex(
                "E550CD532B881D52AD75CE7B91171063E568F2531FBDFB32EE01D1910BCF810F"
            ),
            peers={
                bytes.fromhex("0A"): bytes.fromhex(
                    "CE616F28426EF24EDB51DBCEF7A23305F886F657959D4DF889DDFC0255042159"
                ),
                bytes.fromhex("51"): bytes.fromhex(
                    "2668BA6CA302F14E952228DA1250A890C143FDBA4DAED27246188B9E42C94B6D"
                ),
                # that's a new one for deterministic, and we build a key for it but no shared secret
                bytes.fromhex("dc"): None,
            },
            # requests in group mode would be sent with
            #                    ).pairwise_for(bytes.fromhex('0A'))
            # but here we want to request deterministically
        ).for_sending_deterministic_requests(bytes.fromhex("dc"), bytes.fromhex("0a"))
    )

    request = Message(code=GET, uri="coap://localhost/.well-known/core")
    response = await protocol.request(request).response
    print("Got response", response, response.payload)

    request = Message(code=GET, uri="coap://localhost/.well-known/core")
    response = await protocol.request(request).response
    print("Got response", response, response.payload)


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""A tool for creating key pairs for use in credentials

Note that this tool operates on secrets stored in unencrypted files, protectd
by restrictively set file system permissions. While this is common practice
with many tools in the UNIX world, it might be surprising to users coming from
multi-factor environments."""

import aiocoap.meta
import aiocoap.defaults

import argparse
from pathlib import Path


def build_parser():
    p = argparse.ArgumentParser(description=__doc__)
    p.add_argument(
        "--version", action="version", version="%(prog)s " + aiocoap.meta.version
    )

    subparsers = p.add_subparsers(required=True, dest="subcommand")
    generate = subparsers.add_parser(
        "generate",
        help="This generates an"
        " EDHOC key, stores it in a key file that is only readable by the"
        " user, and prints the corresponding public key information in a way"
        " suitable for inclusion in credentials maps.",
    )
    generate.add_argument("keyfile", help="File to store the secret key in", type=Path)
    generate.add_argument(
        "--kid", help="Hexadecimal key identifier", type=bytes.fromhex
    )
    generate.add_argument("--subject", help="Text placed in the CCS", type=str)

    return p


def main():
    p = build_parser()

    args = p.parse_args()

    missmods = aiocoap.defaults.oscore_missing_modules()
    if missmods:
        p.error(
            f'Dependencies missing, consider installing aiocoap as "aiocoap[oscore]" or "aiocoap[all]". Missing modules: {", ".join(missmods)}'
        )

    from aiocoap import edhoc
    import cbor2
    import cbor_diag

    if args.subcommand == "generate":
        try:
            key = edhoc.CoseKeyForEdhoc.generate(args.keyfile)
        except FileExistsError:
            raise p.error("Output file already exists")

        public = key.as_ccs(args.kid, args.subject)
        print(cbor_diag.cbor2diag(cbor2.dumps(public, canonical=True), pretty=False))
    else:
        raise RuntimeError(f"Unimplemented subcommand {args.subcommand=}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import functools
import re
import os
import textwrap
import glob
import os.path
from pathlib import Path

from docutils.parsers.rst.directives.misc import Include

rtd_re = re.compile(
    "^\\.\\. _([^:]+): http://aiocoap.readthedocs.io/en/latest/(.*)\\.html$"
)
filelink_re = re.compile("^\\.\\. _([^:]+): ([^:]+)$")


def addrepl(replacements, pattern, prefix, linkbody):
    """Given details of a sphinx-ified link, place an entry in the replacements
    dictionary that is suitable for full-text substitution later on"""
    pretty = pattern.strip("`")
    if pretty != linkbody:
        sphinxlink = ":%s:`%s <%s>`" % (prefix, pretty, linkbody)
    else:
        sphinxlink = ":%s:`%s`" % (prefix, linkbody)
    replacements[pattern + "_"] = sphinxlink


def modified_insert_input(include_lines, path, original=None):
    """A filter for the insert_input function that preprocesses the input in
    the following ways:

    * Remove all external hyperlink targets to readthedocs (and guess the way
      this link needs to be expressed in Sphinx)
    * Remove all local (relative) link targets
    * Replace words referencing them with the appropriate Sphinx reference"""
    new_lines = []
    replacements = {}
    for line in include_lines:
        rtd_match = rtd_re.match(line)
        filelink_match = filelink_re.match(line)
        if rtd_match:
            pattern, linkbody = rtd_match.groups()
            prefix = "doc"
            addrepl(replacements, pattern, prefix, linkbody)
        elif filelink_match:
            # for things like LICENSE
            pattern, linkbody = filelink_match.groups()
            addrepl(replacements, pattern, "doc", linkbody)
        else:
            # only keep lines that are still relevant (that's most lines)
            new_lines.append(line)
    new_lines = [
        functools.reduce(lambda s, rep: s.replace(*rep), replacements.items(), x)
        for x in new_lines
    ]
    original(new_lines, path)


class IncludePreprocessed(Include):
    """Include a file (like the 'Include' directive), but preprocess its input
    as described in modified_insert_input."""

    def run(self):
        self.state_machine.insert_input = functools.partial(
            modified_insert_input, original=self.state_machine.insert_input
        )
        try:
            result = super().run()
        finally:
            del self.state_machine.insert_input
        return result


def build_moduledocs(app):
    """Create per-module sources like sphinx-apidoc, but at build time and with
    customizations."""
    # Beware that in sphinx as of Debian bookworm, this is not a Path yet (but
    # docs building is only regularly tested on newer versions)
    srcdir = Path(app.builder.srcdir)

    moddir = srcdir / "module"
    os.makedirs(moddir, exist_ok=True)

    basedir = srcdir.parent
    docs = [
        x.removeprefix("src/")
        .removesuffix(".py")
        .replace("/", ".")
        .replace(".__init__", "")
        for x in glob.glob("src/aiocoap/**/*.py", recursive=True, root_dir=basedir)
    ]

    for x in docs:
        commonstart = textwrap.dedent(f"""\
            {x} module
            ====================================================================================
            """)

        if x in ("aiocoap.numbers", "aiocoap.transports"):
            # They have explicit intros pointing out submodules and/or
            # describing any reexports
            text = commonstart + textwrap.dedent(f"""
                .. automodule:: {x}
                .. toctree::
                    :glob:

                    {x}.*
                """)
        elif x in ("aiocoap",):
            # They have explicit intros listing submodules
            text = commonstart + textwrap.dedent(f"""
                .. automodule:: {x}
                """)
        elif x.startswith("aiocoap.cli."):
            if x in ("aiocoap.cli.defaults", "aiocoap.cli.common"):
                # These neither have a man page, nor do they go into the documentation
                continue
            executablename = "aiocoap-" + x.removeprefix("aiocoap.cli.")
            # no ".. automodule:: {x}" because the doc string is already used
            # by the argparse, and thus would be repeated
            text = textwrap.dedent(f"""
                    {executablename}
                    ==============================

                    .. argparse::
                        :ref: {x}.build_parser
                        :prog: {executablename}

                    """)
        else:
            text = commonstart + textwrap.dedent(f"""
                .. automodule:: {x}
                    :members:
                    :undoc-members:
                    :show-inheritance:
                """)
        docname = f"{moddir}/{x}.rst"

        if os.path.exists(docname) and open(docname).read() == text:
            continue
        else:
            with open(moddir / (x + ".rst"), "w") as outfile:
                outfile.write(text)

    for f in os.listdir(moddir):
        if f.endswith(".rst") and f.removesuffix(".rst") not in docs:
            os.unlink(moddir + "/" + f)


def setup(app):
    """Sphinx extension that builds the aiocoap index page from a non-sphinx
    and thus github-suitable ReST file, and also creates sphinx-apidoc style
    per-module pages customized to what I'd like to see there"""

    app.add_directive("include_preprocessed", IncludePreprocessed)
    app.connect("builder-inited", build_moduledocs)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import unittest
import aiocoap.defaults

oscore_modules = aiocoap.defaults.oscore_missing_modules()
_skip_unless_oscore = unittest.skipIf(
    oscore_modules, "Modules missing for running OSCORE tests: %s" % (oscore_modules,)
)


@_skip_unless_oscore
class UtilCryptographyAdditions(unittest.TestCase):
    def test(self):
        from cryptography.hazmat.primitives.asymmetric import ed25519
        from aiocoap.util.cryptography_additions import (
            sk_to_curve25519,
            pk_to_curve25519,
        )

        # From https://github.com/ace-wg/Hackathon-109/blob/master/GroupKeys.md and
        # https://github.com/ace-wg/Hackathon-109/blob/master/GroupDerivation.md,
        # private1 is Rikard Test 2 Entity 1, private2 is Rikard Test 2 Entity 2
        private1 = bytes.fromhex(
            "397CEB5A8D21D74A9258C20C33FC45AB152B02CF479B2E3081285F77454CF347"
        )
        private2 = bytes.fromhex(
            "70559B9EECDC578D5FC2CA37F9969630029F1592AFF3306392AB15546C6A184A"
        )
        shared_reference = bytes.fromhex(
            "4546babdb9482396c167af11d21953bfa49eb9f630c45de93ee4d3b9ef059576"
        )

        private1 = ed25519.Ed25519PrivateKey.from_private_bytes(private1)
        private2 = ed25519.Ed25519PrivateKey.from_private_bytes(private2)

        private1_x = sk_to_curve25519(private1)
        private2_x = sk_to_curve25519(private2)

        assert (
            private1_x.exchange(private2_x.public_key())
            == private2_x.exchange(private1_x.public_key())
            == shared_reference
        )

        public2 = private2.public_key()
        public2_x = pk_to_curve25519(public2)

        assert private1_x.exchange(public2_x) == shared_reference



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

import sys
import os

# for aiocoap_index.py
sys.path.insert(0, os.path.abspath("."))
# for aiocoap.meta
sys.path.insert(0, os.path.abspath(".."))

import aiocoap.meta

extensions = [
    "sphinx.ext.autodoc",
    "aiocoap_index",
    "sphinxarg.ext",
]

source_suffix = ".rst"

# This is renamed and already has the default "index" in later Sphinx versions,
# but so far readthedocs renders this with Sphinx 1 (this is more easily
# addressed after an update to use pyproject.toml)
master_doc = "index"

project = "aiocoap"
copyright = "Christian Amsüss and the aiocoap contributors"

# The full version, including alpha/beta/rc tags.
release = aiocoap.meta.version
# The short X.Y version.
version = ".".join(release.split(".")[:2])

html_logo = "logo.svg"
html_favicon = "logo-square.svg"

autodoc_member_order = "bysource"

man_pages = [
    (
        "module/aiocoap.cli.client",
        "aiocoap-client",
        "query CoAP servers from the command line",
        "",
        1,
    ),
    (
        "module/aiocoap.cli.proxy",
        "aiocoap-proxy",
        "forward and reverse proxy server for CoAP",
        "",
        1,
    ),
    ("module/aiocoap.cli.rd", "aiocoap-rd", "Resource Directory server", "", 1),
    (
        "module/aiocoap.cli.fileserver",
        "aiocoap-fileserver",
        "File server for CoAP",
        "",
        1,
    ),
]



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""This is a usage example of aiocoap that demonstrates how to implement a
simple client. See the "Usage Examples" section in the aiocoap documentation
for some more information."""

import logging
import asyncio

from aiocoap import *

logging.basicConfig(level=logging.INFO)


async def main():
    protocol = await Context.create_client_context()

    request = Message(code=GET, uri="coap://localhost/time")

    try:
        response = await protocol.request(request).response
    except Exception as e:
        print("Failed to fetch resource:")
        print(e)
    else:
        print("Result: %s\n%r" % (response.code, response.payload))


if __name__ == "__main__":
    asyncio.run(main())



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""Confront an aiocoap server with a client that speaks so bad protocol it is
easier to mock with sending byte sequences than with aiocoap"""

import socket
import asyncio
from asyncio import wait_for, TimeoutError
import os
import unittest

import aiocoap

from .fixtures import precise_warnings, no_warnings, asynctest
from .test_server import WithTestServer, WithAsyncLoop

# For some reasons site-local requests do not work on my test setup, resorting
# to link-local; that means a link needs to be given, and while we never need
# to find the default multicast interface to join MC groups, we need to know it
# to address them. This needs support from outside the test suite right now.
_skip_unless_defaultmcif = unittest.skipIf(
    "AIOCOAP_TEST_MCIF" not in os.environ,
    "Multicast tests require AIOCOAP_TEST_MCIF environment variable to tell"
    " the default multicast interface",
)


class MockSockProtocol:
    def __init__(self, remote_addr):
        # It should be pointed out here that this whole mocksock thing is not
        # terribly well thought out, and just hacked together to replace the
        # blocking sockets that used to be there (which were equally hacked
        # together)

        self.incoming_queue = asyncio.Queue()
        self.remote_addr = remote_addr

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, addr):
        self.incoming_queue.put_nowait((data, addr))

    async def close(self):
        pass

    # emulating the possibly connected socket.socket this once was

    def send(self, data):
        self.transport.sendto(data, self.remote_addr)

    def sendto(self, data, addr):
        self.transport.sendto(data, addr)

    async def recv(self):
        return (await self.incoming_queue.get())[0]


class WithMockSock(WithAsyncLoop):
    def setUp(self):
        super().setUp()
        _, self.mocksock = self.loop.run_until_complete(
            self.loop.create_datagram_endpoint(
                lambda: MockSockProtocol(self.mocksock_remote_addr),
                family=socket.AF_INET6,
            )
        )

    def tearDown(self):
        self.loop.run_until_complete(self.mocksock.close())

        super().tearDown()


class TestNoncoapClient(WithTestServer, WithMockSock):
    def setUp(self):
        self.mocksock_remote_addr = (self.serveraddress, aiocoap.COAP_PORT)

        super().setUp()

    @precise_warnings(["Ignoring unparsable message from ..."])
    @asynctest
    async def test_veryshort(self):
        self.mocksock.send(b"\x40")
        await asyncio.sleep(0.1)

    @precise_warnings(["Ignoring unparsable message from ..."])
    @asynctest
    async def test_short_mid(self):
        self.mocksock.send(b"\x40\x01\x97")
        await asyncio.sleep(0.1)

    @precise_warnings(["Ignoring unparsable message from ..."])
    @asynctest
    async def test_version2(self):
        self.mocksock.send(b"\x80\x01\x99\x98")
        await asyncio.sleep(0.1)

    @no_warnings
    @asynctest
    async def test_duplicate(self):
        self.mocksock.send(b"\x40\x01\x99\x99")  # that's a GET /
        await asyncio.sleep(0.1)
        self.mocksock.send(b"\x40\x01\x99\x99")  # that's a GET /
        await asyncio.sleep(0.1)
        r1 = r2 = None
        try:
            r1 = await wait_for(self.mocksock.recv(), timeout=1)
            r2 = await wait_for(self.mocksock.recv(), timeout=1)
        except TimeoutError:
            pass
        self.assertEqual(r1, r2, "Duplicate GETs gave different responses")
        self.assertTrue(r1 is not None, "No responses received to duplicate GET")

    @no_warnings
    @asynctest
    async def test_ping(self):
        self.mocksock.send(
            b"\x40\x00\x99\x9a"
        )  # CoAP ping -- should this test be doable in aiocoap?
        response = await asyncio.wait_for(self.mocksock.recv(), timeout=1)
        assert response == b"\x70\x00\x99\x9a"

    @no_warnings
    @asynctest
    async def test_noresponse(self):
        self.mocksock.send(
            b"\x50\x01\x99\x9b\xd1\xf5\x02"
        )  # CoAP NON GET / with no-response on 2.xx
        try:
            response = await wait_for(self.mocksock.recv(), timeout=1)
            self.assertTrue(
                False, "Response was sent when No-Response should have suppressed it"
            )
        except TimeoutError:
            pass

    @no_warnings
    @asynctest
    async def test_unknownresponse_reset(self):
        self.mocksock.send(
            bytes.fromhex("4040ffff")
        )  # CoAP CON 2.00 that the server has not sent a request for
        response = await wait_for(self.mocksock.recv(), timeout=1)
        self.assertEqual(
            response,
            bytes.fromhex("7000ffff"),
            "Unknown CON Response did not trigger RST",
        )

    @no_warnings
    @asynctest
    async def test_unknownresponse_noreset(self):
        self.mocksock.send(
            bytes.fromhex("6040ffff")
        )  # CoAP ACK 2.00 that the server has not sent a request for
        try:
            response = await wait_for(self.mocksock.recv(), timeout=1)
            self.assertTrue(False, "Unknown ACK Response triggered something")
        except TimeoutError:
            pass


# Skipping the whole class when no multicast address was given (as otherwise
# it'd try binding :: which is bound to fail with a simplesocketserver setting)
@_skip_unless_defaultmcif
class TestNoncoapMulticastClient(WithTestServer, WithMockSock):
    # This exposes the test server to traffic from the environment system for
    # some time; it's only run if a default multicast inteface is given
    # explicitly, though.
    serveraddress = "::"

    def setUp(self):
        # always used with sendto
        self.mocksock_remote_addr = None

        super().setUp()

    @no_warnings
    @asynctest
    async def test_mutlicast_ping(self):
        # exactly like the unicast case -- just to verify we're actually reaching our server
        self.mocksock.sendto(
            b"\x40\x00\x99\x9a",
            (
                aiocoap.numbers.constants.MCAST_IPV6_LINKLOCAL_ALLCOAPNODES,
                aiocoap.COAP_PORT,
                0,
                socket.if_nametoindex(os.environ["AIOCOAP_TEST_MCIF"]),
            ),
        )
        response = await wait_for(self.mocksock.recv(), timeout=1)
        assert response == b"\x70\x00\x99\x9a"

    @no_warnings
    @asynctest
    async def test_multicast_unknownresponse_noreset(self):
        self.mocksock.sendto(
            bytes.fromhex("4040ffff"),
            (
                aiocoap.numbers.constants.MCAST_IPV6_LINKLOCAL_ALLCOAPNODES,
                aiocoap.COAP_PORT,
                0,
                socket.if_nametoindex(os.environ["AIOCOAP_TEST_MCIF"]),
            ),
        )
        try:
            response = await wait_for(self.mocksock.recv(), timeout=1)
        except TimeoutError:
            pass
        else:
            self.assertEqual(
                False,
                "Message was sent back responding to CON response to multicast address",
            )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

from setuptools import setup, Command


class Cite(Command):
    description = """Print how to cite aiocoap in a publication"""

    user_options = [("bibtex", None, "Output citation data as bibtex")]
    boolean_options = ["bibtex"]

    def initialize_options(self):
        self.bibtex = False

    def finalize_options(self):
        pass

    def run(self):
        if self.bibtex:
            print(self.bibtex_text)
        else:
            print(self.plain_text)

    plain_text = """Amsüss, Christian and the aiocoap contributors. aiocoap: Python CoAP Library. 2013–. https://christian.amsuess.com/tools/aiocoap/"""

    bibtex_text = """@Misc{,
        author = {Christian Amsüss and aiocoap contributors},
        title = {{aiocoap}: Python CoAP Library},
        year = {2013--},
        url = {https://christian.amsuess.com/tools/aiocoap/},
        }"""


setup(
    cmdclass={
        "cite": Cite,
    },
)



# SPDX-FileCopyrightText: Christian Amsüss and the aiocoap contributors
#
# SPDX-License-Identifier: MIT

"""a plain CoAP proxy that can work both as forward and as reverse proxy"""

import sys
import argparse

import aiocoap
from aiocoap.proxy.server import (
    ForwardProxyWithPooledObservations,
    ProxyWithPooledObservations,
    NameBasedVirtualHost,
    SubdomainVirtualHost,
    SubresourceVirtualHost,
    UnconditionalRedirector,
)
from aiocoap.util.cli import AsyncCLIDaemon
from aiocoap.cli.common import add_server_arguments, server_context_from_arguments


def build_parser():
    p = argparse.ArgumentParser(description=__doc__)

    mode = p.add_argument_group(
        "mode", "Required argument for setting the operation mode"
    )
    mode.add_argument("--forward", help="Run as forward proxy", action="store_true")
    mode.add_argument("--reverse", help="Run as reverse proxy", action="store_true")

    details = p.add_argument_group(
        "details", "Options that govern how requests go in and out"
    )
    add_server_arguments(details)
    details.add_argument(
        "--register",
        help="Register with a Resource directory",
        metavar="RD-URI",
        nargs="?",
        default=False,
    )
    details.add_argument(
        "--register-as",
        help="Endpoint name (with possibly a domain after a dot) to register as",
        metavar="EP[.D]",
        default=None,
    )
    details.add_argument(
        "--register-proxy",
        help="Ask the RD to serve as a reverse proxy. Note that this is only practical for --unconditional or --pathbased reverse proxies.",
        action="store_true",
    )

    r = p.add_argument_group(
        "Rules",
        description="Sequence of forwarding rules "
        "that, if matched by a request, specify a forwarding destination. Destinations can be prefixed to change their behavior: With an '@' sign, they are treated as forward proxies. With a '!' sign, the destination is set as Uri-Host.",
    )

    class TypedAppend(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            if getattr(namespace, self.dest) is None:
                setattr(namespace, self.dest, [])
            getattr(namespace, self.dest).append((option_string, values))

    r.add_argument(
        "--namebased",
        help="If Uri-Host matches NAME, route to DEST",
        metavar="NAME:DEST",
        action=TypedAppend,
        dest="r",
    )
    r.add_argument(
        "--subdomainbased",
        help="If Uri-Host is anything.NAME, route to DEST",
        metavar="NAME:DEST",
        action=TypedAppend,
        dest="r",
    )
    r.add_argument(
        "--pathbased",
        help="If a requested path starts with PATH, split that part off and route to DEST",
        metavar="PATH:DEST",
        action=TypedAppend,
        dest="r",
    )
    r.add_argument(
        "--unconditional",
        help="Route all requests not previously matched to DEST",
        metavar="DEST",
        action=TypedAppend,
        dest="r",
    )

    return p


def destsplit(dest):
    use_as_proxy = False
    rewrite_uri_host = False
    if dest.startswith("!"):
        dest = dest[1:]
        rewrite_uri_host = True
    if dest.startswith("@"):
        dest = dest[1:]
        use_as_proxy = True
    return dest, rewrite_uri_host, use_as_proxy


class Main(AsyncCLIDaemon):
    async def start(self, args=None):
        parser = build_parser()
        options = parser.parse_args(args if args is not None else sys.argv[1:])
        self.options = options

        if not options.forward and not options.reverse:
            raise parser.error("At least one of --forward and --reverse must be given.")

        self.outgoing_context = await aiocoap.Context.create_client_context()
        if options.forward:
            proxy = ForwardProxyWithPooledObservations(self.outgoing_context)
        else:
            proxy = ProxyWithPooledObservations(self.outgoing_context)
        for kind, data in options.r or ():
            if kind in ("--namebased", "--subdomainbased"):
                try:
                    name, dest = data.split(":", 1)
                except Exception:
                    raise parser.error("%s needs NAME:DEST as arguments" % kind)
                dest, rewrite_uri_host, use_as_proxy = destsplit(dest)
                if rewrite_uri_host and kind == "--subdomainbased":
                    parser.error(
                        "The flag '!' makes no sense for subdomain based redirection as the subdomain data would be lost"
                    )
                r = (
                    NameBasedVirtualHost
                    if kind == "--namebased"
                    else SubdomainVirtualHost
                )(name, dest, rewrite_uri_host, use_as_proxy)
            elif kind == "--pathbased":
                try:
                    path, dest = data.split(":", 1)
                except Exception:
                    raise parser.error("--pathbased needs PATH:DEST as arguments")
                r = SubresourceVirtualHost(path.split("/"), dest)
            elif kind == "--unconditional":
                dest, rewrite_uri_host, use_as_proxy = destsplit(data)
                if rewrite_uri_host:
                    parser.error(
                        "The flag '!' makes no sense for unconditional redirection as the host name data would be lost"
                    )
                r = UnconditionalRedirector(dest, use_as_proxy)
            else:
                raise AssertionError("Unknown redirectory kind")
            proxy.add_redirector(r)

        self.proxy_context = await server_context_from_arguments(proxy, options)

        if options.register is not False:
            from aiocoap.resourcedirectory.client.register import Registerer

            params = {}
            if options.register_as:
                ep, _, d = options.register_as.partition(".")
                params["ep"] = ep
                if d:
                    params["d"] = d
            if options.register_proxy:
                # FIXME: Check this in discovery
                params["proxy"] = "on"
            # FIXME: Construct this from settings (path-based), and forward results
            proxy.get_resources_as_linkheader = lambda: ""
            self.registerer = Registerer(
                self.proxy_context,
                rd=options.register,
                lt=60,
                registration_parameters=params,
            )

    async def shutdown(self):
        if self.options.register is not False:
            await self.registerer.shutdown()
        await self.outgoing_context.shutdown()
        await self.proxy_context.shutdown()


sync_main = Main.sync_main

if __name__ == "__main__":
    # if you want to run this using `python3 -m`, see http://bugs.python.org/issue22480
    sync_main()
